"""Report data namespace

Return data optimized for reports implemented in the Reporter
blueprint for consumption by external systems.
"""

from flask import request
from flask_restx import Namespace
from flask_restx import Resource

from AIPscan.API import fields
from AIPscan.Data import report_data
from AIPscan.helpers import parse_bool
from AIPscan.helpers import parse_datetime_bound

api = Namespace(
    "report-data", description="Retrieve data optimized for AIPscan reports"
)


@api.route("/format-version-count/<storage_service_id>")
class FormatVersionList(Resource):
    @api.doc(
        "list_format_versions",
        params={
            fields.FIELD_START_DATE: {
                "description": "AIP creation start date (inclusive, YYYY-MM-DD)",
                "in": "query",
                "type": "str",
            },
            fields.FIELD_END_DATE: {
                "description": "AIP creation end date (inclusive, YYYY-MM-DD)",
                "in": "query",
                "type": "str",
            },
            fields.FIELD_STORAGE_LOCATION: {
                "description": "Storage Location ID",
                "in": "query",
                "type": "int",
            },
        },
    )
    def get(self, storage_service_id):
        """List file format versions with file count and size"""
        start_date = parse_datetime_bound(request.args.get(fields.FIELD_START_DATE))
        end_date = parse_datetime_bound(
            request.args.get(fields.FIELD_END_DATE), upper=True
        )
        storage_location_id = request.args.get(fields.FIELD_STORAGE_LOCATION)

        return report_data.format_versions_count(
            storage_service_id=storage_service_id,
            start_date=start_date,
            end_date=end_date,
            storage_location_id=storage_location_id,
        )


@api.route("/largest-aips/<storage_service_id>")
class LargestAIPList(Resource):
    @api.doc(
        "list_largest_aips",
        params={
            fields.FIELD_START_DATE: {
                "description": "AIP creation start date (inclusive, YYYY-MM-DD)",
                "in": "query",
                "type": "str",
            },
            fields.FIELD_END_DATE: {
                "description": "AIP creation end date (inclusive, YYYY-MM-DD)",
                "in": "query",
                "type": "str",
            },
            fields.FIELD_LIMIT: {
                "description": "Number of results to return (default is 20)",
                "in": "query",
                "type": "int",
            },
            fields.FIELD_STORAGE_LOCATION: {
                "description": "Storage Location ID",
                "in": "query",
                "type": "int",
            },
        },
    )
    def get(self, storage_service_id, limit=20):
        """List largest AIPs"""
        start_date = parse_datetime_bound(request.args.get(fields.FIELD_START_DATE))
        end_date = parse_datetime_bound(
            request.args.get(fields.FIELD_END_DATE), upper=True
        )
        storage_location_id = request.args.get(fields.FIELD_STORAGE_LOCATION)
        try:
            limit = int(request.args.get(fields.FIELD_LIMIT, 20))
        except ValueError:
            limit = 20
        return report_data.largest_aips(
            storage_service_id=storage_service_id,
            start_date=start_date,
            end_date=end_date,
            storage_location_id=storage_location_id,
            limit=limit,
        )


@api.route("/largest-files/<storage_service_id>")
class LargestFileList(Resource):
    @api.doc(
        "list_largest_files",
        params={
            fields.FIELD_START_DATE: {
                "description": "AIP creation start date (inclusive, YYYY-MM-DD)",
                "in": "query",
                "type": "str",
            },
            fields.FIELD_END_DATE: {
                "description": "AIP creation end date (inclusive, YYYY-MM-DD)",
                "in": "query",
                "type": "str",
            },
            fields.FIELD_FILE_TYPE: {
                "description": "Optional file type filter (original or preservation)",
                "in": "query",
                "type": "str",
            },
            fields.FIELD_LIMIT: {
                "description": "Number of results to return (default is 20)",
                "in": "query",
                "type": "int",
            },
            fields.FIELD_STORAGE_LOCATION: {
                "description": "Storage Location ID",
                "in": "query",
                "type": "int",
            },
        },
    )
    def get(self, storage_service_id, file_type=None, limit=20):
        """List largest files"""
        start_date = parse_datetime_bound(request.args.get(fields.FIELD_START_DATE))
        end_date = parse_datetime_bound(
            request.args.get(fields.FIELD_END_DATE), upper=True
        )
        file_type = request.args.get(fields.FIELD_FILE_TYPE)
        storage_location_id = request.args.get(fields.FIELD_STORAGE_LOCATION)
        try:
            limit = int(request.args.get(fields.FIELD_LIMIT, 20))
        except ValueError:
            limit = 20
        return report_data.largest_files(
            storage_service_id=storage_service_id,
            start_date=start_date,
            end_date=end_date,
            storage_location_id=storage_location_id,
            file_type=file_type,
            limit=limit,
        )


@api.route("/aips-by-file-format/<storage_service_id>")
class AIPsByFormatList(Resource):
    @api.doc(
        "list_aips_by_format",
        params={
            fields.FIELD_FILE_FORMAT: {
                "description": "File format name (must be exact match)",
                "in": "query",
                "type": "str",
            },
            fields.FIELD_ORIGINAL_FILES: {
                "description": "Return data for original files (default) or preservation derivatives",
                "in": "query",
                "type": "bool",
            },
            fields.FIELD_STORAGE_LOCATION: {
                "description": "Storage Location ID",
                "in": "query",
                "type": "int",
            },
        },
    )
    def get(self, storage_service_id):
        """List AIPs containing given file format"""
        file_format = request.args.get(fields.FIELD_FILE_FORMAT, "")
        original_files = parse_bool(request.args.get(fields.FIELD_ORIGINAL_FILES, True))
        storage_location_id = request.args.get(fields.FIELD_STORAGE_LOCATION)
        return report_data.aips_by_file_format(
            storage_service_id=storage_service_id,
            file_format=file_format,
            original_files=original_files,
            storage_location_id=storage_location_id,
        )


@api.route("/aips-by-puid/<storage_service_id>")
class AIPsByPUIDList(Resource):
    @api.doc(
        "list_aips_by_puid",
        params={
            fields.FIELD_PUID: {
                "description": "PRONOM ID (PUID)",
                "in": "query",
                "type": "str",
            },
            fields.FIELD_ORIGINAL_FILES: {
                "description": "Return data for original files (default) or preservation derivatives",
                "in": "query",
                "type": "bool",
            },
            fields.FIELD_STORAGE_LOCATION: {
                "description": "Storage Location ID",
                "in": "query",
                "type": "int",
            },
        },
    )
    def get(self, storage_service_id):
        """List AIPs containing given format version, specified by PUID"""
        puid = request.args.get(fields.FIELD_PUID, "")
        original_files = parse_bool(request.args.get(fields.FIELD_ORIGINAL_FILES, True))
        storage_location_id = request.args.get(fields.FIELD_STORAGE_LOCATION)
        return report_data.aips_by_puid(
            storage_service_id=storage_service_id,
            puid=puid,
            original_files=original_files,
            storage_location_id=storage_location_id,
        )


@api.route("/agents-transfers/<storage_service_id>")
class AgentData(Resource):
    @api.doc(
        "agent_info",
        params={
            fields.FIELD_START_DATE: {
                "description": "AIP creation start date (inclusive, YYYY-MM-DD)",
                "in": "query",
                "type": "str",
            },
            fields.FIELD_END_DATE: {
                "description": "AIP creation end date (inclusive, YYYY-MM-DD)",
                "in": "query",
                "type": "str",
            },
            fields.FIELD_STORAGE_LOCATION: {
                "description": "Storage Location ID",
                "in": "query",
                "type": "int",
            },
        },
    )
    def get(self, storage_service_id):
        """List user agents and their transfers"""
        start_date = parse_datetime_bound(request.args.get(fields.FIELD_START_DATE))
        end_date = parse_datetime_bound(
            request.args.get(fields.FIELD_END_DATE), upper=True
        )
        storage_location_id = request.args.get(fields.FIELD_STORAGE_LOCATION)
        return report_data.agents_transfers(
            storage_service_id=storage_service_id,
            start_date=start_date,
            end_date=end_date,
            storage_location_id=storage_location_id,
        )


@api.route("/storage_locations/<storage_service_id>")
class StorageLocations(Resource):
    @api.doc(
        "storage_locations",
        params={
            fields.FIELD_START_DATE: {
                "description": "AIP creation start date (inclusive, YYYY-MM-DD)",
                "in": "query",
                "type": "str",
            },
            fields.FIELD_END_DATE: {
                "description": "AIP creation end date (inclusive, YYYY-MM-DD)",
                "in": "query",
                "type": "str",
            },
        },
    )
    def get(self, storage_service_id):
        """List AIP store locations and their usage."""
        start_date = parse_datetime_bound(request.args.get(fields.FIELD_START_DATE))
        end_date = parse_datetime_bound(
            request.args.get(fields.FIELD_END_DATE), upper=True
        )
        return report_data.storage_locations(
            storage_service_id=storage_service_id,
            start_date=start_date,
            end_date=end_date,
        )


@api.route("/storage_location_usage_over_time/<storage_service_id>")
class StorageLocationsUsageOverTime(Resource):
    @api.doc(
        "storage_locations_usage_over_time",
        params={
            fields.FIELD_START_DATE: {
                "description": "AIP creation start date (inclusive, YYYY-MM-DD)",
                "in": "query",
                "type": "str",
            },
            fields.FIELD_END_DATE: {
                "description": "AIP creation end date (inclusive, YYYY-MM-DD)",
                "in": "query",
                "type": "str",
            },
            fields.FIELD_CUMULATIVE: {
                "description": "Return per month storage location usage: differential (False, default) or cumulative (True)",
                "in": "query",
                "type": "bool",
            },
        },
    )
    def get(self, storage_service_id):
        """List AIP store locations and their usage."""
        start_date = parse_datetime_bound(request.args.get(fields.FIELD_START_DATE))
        end_date = parse_datetime_bound(
            request.args.get(fields.FIELD_END_DATE), upper=True
        )
        cumulative = parse_bool(
            request.args.get(fields.FIELD_CUMULATIVE), default=False
        )
        return report_data.storage_locations_usage_over_time(
            storage_service_id=storage_service_id,
            start_date=start_date,
            end_date=end_date,
            cumulative=cumulative,
        )



"""Functions to help us tease apart a METS file and write to the
database.
"""

import os
from datetime import datetime

from celery.utils.log import get_task_logger
from lxml import etree

from AIPscan import db
from AIPscan.Aggregator import tasks
from AIPscan.Aggregator.downloads import get_download_root
from AIPscan.Aggregator.task_helpers import _tz_neutral_date
from AIPscan.Aggregator.task_helpers import get_storage_service_api_url
from AIPscan.models import AIP
from AIPscan.models import Agent
from AIPscan.models import Event
from AIPscan.models import EventAgent
from AIPscan.models import FetchJob
from AIPscan.models import File
from AIPscan.models import FileType
from AIPscan.models import Pipeline
from AIPscan.models import StorageLocation

logger = get_task_logger(__name__)


ORIGINAL_OBJECT = "original"
PRESERVATION_OBJECT = "preservation"


def _extract_event_detail(premis_event, file_id):
    """Extract the detail from the event and write a new event object
    to the database"""
    event_type = premis_event.event_type
    event_uuid = premis_event.event_identifier_value
    event_date = _tz_neutral_date(premis_event.event_date_time)
    # We have a strange issue with this logged: https://github.com/archivematica/Issues/issues/743
    event_detail, event_outcome, event_outcome_detail = None, None, None
    if not isinstance(premis_event.event_detail, tuple):
        event_detail = premis_event.event_detail
    if not isinstance(premis_event.event_outcome, tuple):
        event_outcome = premis_event.event_outcome
    if not isinstance(premis_event.event_outcome_detail_note, tuple):
        event_outcome_detail = premis_event.event_outcome_detail_note
    event = Event(
        event_type=event_type,
        uuid=event_uuid,
        date=event_date,
        detail=event_detail,
        outcome=event_outcome,
        outcome_detail=event_outcome_detail,
        file_id=file_id,
    )
    return event


def _create_agent_type_id(identifier_type, identifier_value):
    """Create a key-pair string for the linking_type_value in the db."""
    return f"{identifier_type}-{identifier_value}"


def _create_event_agent_relationship(event_id, agent_identifier):
    """Generator object helper for looping through an event's agents and
    returning the event-agent IDs.
    """
    for agent_ in agent_identifier:
        id_ = _create_agent_type_id(
            agent_.linking_agent_identifier_type, agent_.linking_agent_identifier_value
        )
        existing_agent = Agent.query.filter_by(linking_type_value=id_).first()
        event_relationship = EventAgent.insert().values(
            event_id=event_id, agent_id=existing_agent.id
        )
        yield event_relationship


def create_event_objects(fs_entry, file_id):
    """Add information about PREMIS Events associated with file to database

    :param fs_entry: mets-reader-writer FSEntry object
    :param file_id: File ID
    """
    for premis_event in fs_entry.get_premis_events():
        event = _extract_event_detail(premis_event, file_id)
        db.session.add(event)
        db.session.commit()

        for event_relationship in _create_event_agent_relationship(
            event.id, premis_event.linking_agent_identifier
        ):
            db.session.execute(event_relationship)
        db.session.commit()


def _extract_agent_detail(agent, storage_service_id):
    """Pull the agent information from the agent record and return an
    agent object ready to insert into the database.
    """
    linking_type_value = agent[0]
    agent_type = agent[1]
    agent_value = agent[2]
    return Agent(
        linking_type_value=linking_type_value,
        agent_type=agent_type,
        agent_value=agent_value,
        storage_service_id=storage_service_id,
    )


def create_agent_objects(unique_agents, storage_service_id):
    """Add our agents to the database. The list is already the
    equivalent of a set by the time it reaches here and so we don't
    need to perform any de-duplication.
    """
    for agent in unique_agents:
        agent_obj = _extract_agent_detail(agent, storage_service_id)
        exists = Agent.query.filter_by(
            linking_type_value=agent_obj.linking_type_value,
            agent_type=agent_obj.agent_type,
            agent_value=agent_obj.agent_value,
            storage_service_id=storage_service_id,
        ).count()
        if exists:
            continue
        logger.info("Adding: %s", agent_obj)
        db.session.add(agent_obj)
    db.session.commit()


def _get_unique_agents(all_agents, agents_list):
    """Return unique agents associated with a file in the AIP. Returns
    a list of tuples organized by (linking_type_value, type,
    value), e.g. (Archivematica user pk 4, Archivematica User, Eric).
    Linking type and value become one because only together are these
    two considered useful, they just happen to exist as two elements
    in the PREMIS dictionary.
    """
    agents = []
    linking_type_value = ""
    for agent in all_agents:
        linking_type_value = _create_agent_type_id(
            agent.agent_identifier[0].agent_identifier_type,
            agent.agent_identifier[0].agent_identifier_value,
        )
        agent_tuple = (linking_type_value, agent.type, agent.name)
        if agent_tuple not in agents_list:
            agents.append(agent_tuple)
    return agents


def create_aip_object(
    package_uuid,
    transfer_name,
    create_date,
    mets_sha256,
    size=0,
    storage_service_id=1,
    storage_location_id=1,
    fetch_job_id=1,
    origin_pipeline_id=1,
):
    """Create an AIP object and save it to the database."""
    aip = AIP(
        uuid=package_uuid,
        transfer_name=transfer_name,
        create_date=_tz_neutral_date(create_date),
        mets_sha256=mets_sha256,
        size=size,
        storage_service_id=storage_service_id,
        storage_location_id=storage_location_id,
        fetch_job_id=fetch_job_id,
        origin_pipeline_id=origin_pipeline_id,
    )
    db.session.add(aip)
    db.session.commit()
    return aip


def delete_aip_object(aip):
    """Delete AIP object from database.

    :param aip: AIP model instance
    """
    db.session.delete(aip)
    db.session.commit()


def create_storage_location_object(current_location, description, storage_service_id):
    """Create a StorageLocation and save it to the database."""
    storage_location = StorageLocation(
        current_location=current_location,
        description=description,
        storage_service_id=storage_service_id,
    )
    db.session.add(storage_location)
    db.session.commit()
    return storage_location


def create_or_update_storage_location(current_location, storage_service):
    """Create or update Storage Location and return it."""
    storage_location = StorageLocation.query.filter_by(
        current_location=current_location
    ).first()

    request_url, request_url_without_api_key = get_storage_service_api_url(
        storage_service, current_location
    )
    response = tasks.make_request(request_url, request_url_without_api_key)
    description = response.get("description")
    if not storage_location:
        return create_storage_location_object(
            current_location, description, storage_service.id
        )

    if storage_location.description != description:
        storage_location.description = description
        db.session.commit()

    return storage_location


def create_pipeline_object(origin_pipeline, dashboard_url):
    """Create a Pipeline and save it to the database."""
    pipeline = Pipeline(origin_pipeline=origin_pipeline, dashboard_url=dashboard_url)
    db.session.add(pipeline)
    db.session.commit()
    return pipeline


def create_or_update_pipeline(origin_pipeline, storage_service):
    """Create or update Storage Location and return it."""
    pipeline = Pipeline.query.filter_by(origin_pipeline=origin_pipeline).first()

    request_url, request_url_without_api_key = get_storage_service_api_url(
        storage_service, origin_pipeline
    )
    response = tasks.make_request(request_url, request_url_without_api_key)
    dashboard_url = response.get("remote_name")
    if not pipeline:
        return create_pipeline_object(origin_pipeline, dashboard_url)

    if pipeline.dashboard_url != dashboard_url:
        pipeline.dashboard_url = dashboard_url
        db.session.commit()

    return pipeline


def _get_file_properties(fs_entry):
    """Retrieve file properties from FSEntry

    :param fs_entry: mets-reader-writer FSEntry object

    :returns: Dict of file properties
    """
    file_info = {
        "uuid": fs_entry.file_uuid,
        "name": fs_entry.label,
        "filepath": fs_entry.path,
        "size": None,
        "date_created": None,
        "puid": None,
        "file_format": None,
        "format_version": None,
        "checksum_type": None,
        "checksum_value": None,
        "related_uuid": None,
    }

    try:
        for premis_object in fs_entry.get_premis_objects():
            file_info["size"] = premis_object.size
            key_alias = premis_object.format_registry_key
            file_info["date_created"] = _tz_neutral_date(
                premis_object.date_created_by_application
            )
            if not isinstance(key_alias, tuple):
                file_info["puid"] = key_alias
            file_info["file_format"] = premis_object.format_name
            version_alias = premis_object.format_version
            if not isinstance(version_alias, tuple):
                file_info["format_version"] = version_alias
            file_info["checksum_type"] = premis_object.message_digest_algorithm
            file_info["checksum_value"] = premis_object.message_digest
            related_uuid_alias = premis_object.related_object_identifier_value
            if not isinstance(related_uuid_alias, tuple):
                file_info["related_uuid"] = related_uuid_alias
    except AttributeError:
        # File error/warning to log. Obviously this format may
        # be incorrect so it is our best guess.
        file_info["file_format"] = "ISO Disk Image File"
        file_info["puid"] = "fmt/468"

    return file_info


def _add_normalization_date(file_id):
    """Add normalization date from PREMIS creation Event to preservation file

    :param file_id: File ID
    """
    file_ = db.session.get(File, file_id)
    creation_event = Event.query.filter_by(file_id=file_.id, type="creation").first()
    if creation_event is not None:
        file_.date_created = creation_event.date
        db.session.commit()


def _add_premis_object_xml(fs_entry, file_id):
    """Add string representation of PREMIS Object to File object."""
    file_ = db.session.get(File, file_id)

    if hasattr(fs_entry, "amdsecs"):
        for ss in fs_entry.amdsecs[0].subsections:
            if ss.contents.mdtype == fs_entry.PREMIS_OBJECT:
                file_.premis_object = etree.tostring(
                    ss.contents.serialize(), encoding="unicode", pretty_print=True
                )
        db.session.commit()


def _get_original_file(related_uuid):
    """Get original file related to preservation derivative."""
    return File.query.filter_by(uuid=related_uuid, file_type=FileType.original).first()


def create_file_object(file_type, fs_entry, aip_id):
    """Add file to database

    :param file_type: models.FileType enum
    :param fs_entry: mets-reader-writer FSEntry object
    :param aip_id: AIP ID
    """
    file_info = _get_file_properties(fs_entry)

    original_file_id = None
    if file_type is FileType.preservation:
        original_file = _get_original_file(file_info["related_uuid"])
        if original_file:
            original_file_id = original_file.id

    new_file = File(
        name=file_info.get("name"),
        filepath=file_info.get("filepath"),
        uuid=file_info.get("uuid"),
        file_type=file_type,
        size=file_info.get("size"),
        date_created=file_info.get("date_created"),
        puid=file_info.get("puid"),
        file_format=file_info.get("file_format"),
        format_version=file_info.get("format_version"),
        checksum_type=file_info.get("checksum_type"),
        checksum_value=file_info.get("checksum_value"),
        original_file_id=original_file_id,
        aip_id=aip_id,
    )

    logger.debug("Adding file %s %s", new_file.name, aip_id)

    db.session.add(new_file)
    db.session.commit()

    create_event_objects(fs_entry, new_file.id)

    if file_type == FileType.preservation:
        _add_normalization_date(new_file.id)

    _add_premis_object_xml(fs_entry, new_file.id)


def collect_mets_agents(mets):
    """Collect all of the unique agents in the METS file to write to the
    database.
    """
    agents = []
    for aip_file in mets.all_files():
        if aip_file.use != ORIGINAL_OBJECT and aip_file.use != PRESERVATION_OBJECT:
            continue
        agents = agents + _get_unique_agents(aip_file.get_premis_agents(), agents)
    logger.info("Total agents: %d", len(agents))
    return agents


def process_aip_data(aip, mets):
    """Populate database with information needed for reporting from METS file

    :param aip: AIP object
    :param mets: mets-reader-writer METSDocument object
    """
    tasks.get_mets.update_state(state="IN PROGRESS")

    create_agent_objects(collect_mets_agents(mets), aip.storage_service_id)

    all_files = mets.all_files()

    # Parse the original files first so that they are available as foreign keys
    # when we parse preservation and derivative files.
    original_files = [file_ for file_ in all_files if file_.use == "original"]
    for file_ in original_files:
        create_file_object(FileType.original, file_, aip.id)

    preservation_files = [file_ for file_ in all_files if file_.use == "preservation"]
    for file_ in preservation_files:
        create_file_object(FileType.preservation, file_, aip.id)


def create_fetch_job(datetime_obj_start, timestamp_str, storage_server_id):
    download_root = get_download_root()
    download_directory = os.path.join(download_root, timestamp_str, "")
    fetch_job = FetchJob(
        total_packages=None,
        total_deleted_aips=None,
        total_aips=None,
        download_start=datetime_obj_start,
        download_end=None,
        download_directory=download_directory,
        storage_service_id=storage_server_id,
    )
    db.session.add(fetch_job)
    db.session.commit()

    return fetch_job


def update_fetch_job(fetch_job_id, processed_packages, total_packages_count):
    # Count different types of packages
    total_aips = 0
    total_sips = 0
    total_dips = 0
    total_deleted_aips = 0
    total_replicas = 0

    for package in processed_packages:
        if package.is_aip():
            total_aips += 1

        if package.is_sip():
            total_sips += 1

        if package.is_dip():
            total_dips += 1

        if package.is_deleted():
            total_deleted_aips += 1

        if package.is_replica():
            total_replicas += 1

    # Store counts of different types of packages
    obj = FetchJob.query.filter_by(id=fetch_job_id).first()
    obj.total_packages = total_packages_count
    obj.total_aips = total_aips
    obj.total_dips = total_dips
    obj.total_sips = total_sips
    obj.total_replicas = total_replicas
    obj.total_deleted_aips = total_deleted_aips
    obj.download_end = datetime.now().replace(microsecond=0)
    db.session.commit()

    return obj



"""Tests for the agents transfer log."""

import pytest
from flask import current_app

from AIPscan.Data import fields
from AIPscan.Reporter.report_ingest_log import get_figure_html
from AIPscan.Reporter.report_ingest_log import get_table_data

EXPECTED_CSV_CONTENTS = b"AIP UUID,AIP Name,Start Date,End Date,User,Duration\r\n111111111111-1111-1111-11111111,Test AIP,2020-12-02 10:00:00,2020-12-02 10:30:32,user one,0:30:32\r\n"

ZERO_TRANSFERS = {"StorageName": None, "Ingests": []}

ONE_TRANSFER = {
    "StorageName": "S1",
    "Ingests": [
        {
            "AIP": "2905e275-90df-4f07-8203-bc54d5588d32",
            "AIPName": "12345",
            "Event": "ingestion",
            "IngestStartDate": "2020-11-30 18:30:00",
            "IngestFinishDate": "2020-11-30 18:35:00",
            "User": "test",
        }
    ],
}

THREE_TRANSFERS = {
    "StorageName": "S2",
    "Ingests": [
        {
            "AIP": "2905e275-90df-4f07-8203-bc54d5588d32",
            "AIPName": "12345",
            "Event": "ingestion",
            "IngestStartDate": "2020-11-30 18:30:00",
            "IngestFinishDate": "2020-11-30 18:30:42",
            "User": "test",
        },
        {
            "AIP": "c8111c6f-d74b-44b6-87dd-668b0fdc59e4",
            "AIPName": "images_two",
            "Event": "ingestion",
            "IngestStartDate": "2020-11-30 18:30:00",
            "IngestFinishDate": "2020-11-30 18:30:42",
            "User": "miller",
        },
        {
            "AIP": "24a5a997-83b6-49bc-b4fb-29ec7a561ce6",
            "AIPName": "longer_transfer",
            "Event": "ingestion",
            "IngestStartDate": "2020-11-30 18:30:00",
            "IngestFinishDate": "2020-11-30 18:30:42",
            "User": "ross",
        },
    ],
}


@pytest.mark.parametrize(
    "agents_transfers, transfer_count, storage_name, duration",
    [
        # There is no data, e.g. the request could be for an invalid
        # storage service. Count should be zero, a valid div is still
        # returned with the data.
        (ZERO_TRANSFERS, 0, None, None),
        # One transfer with valid data. Duration for the first ingest is
        # configured to 5 mins.
        (ONE_TRANSFER, 1, "S1", "0:05:00"),
        # Three transfers with valid data. Duration for the first ingest
        # is configured to 42 seconds.
        (THREE_TRANSFERS, 3, "S2", "0:00:42"),
    ],
)
def test_get_table_data(agents_transfers, transfer_count, storage_name, duration):
    """Test that we get valid data for the transfers log table, e.g.
    including our calculation of duration.
    """
    response = get_table_data(agents_transfers)
    assert response["transfer_count"] == transfer_count
    assert response["StorageName"] == storage_name
    if not duration:
        with pytest.raises(IndexError):
            _ = response["Ingests"][0][fields.FIELD_DURATION]
        return
    assert str(response["Ingests"][0][fields.FIELD_DURATION]) == duration


@pytest.mark.parametrize(
    "agents_transfers, transfer_count, storage_name",
    [
        # There is no data, e.g. the request could be for an invalid
        # storage service. Count should be zero, a valid div is still
        # returned with the data.
        (ZERO_TRANSFERS, 0, None),
        # One transfer with valid data.
        (ONE_TRANSFER, 1, "S1"),
        # Three transfers with valid data.
        (THREE_TRANSFERS, 3, "S2"),
    ],
)
def test_get_figure_html(agents_transfers, transfer_count, storage_name):
    """Test that we get valid data required for the Gantt chart to be
    rendered.
    """
    response = get_figure_html(agents_transfers)
    assert response["transfer_count"] == transfer_count
    assert response["StorageName"] == storage_name
    assert response["figure"].startswith("<div>")
    assert response["figure"].endswith("</div>")


def test_user_ingest_log(app_with_populated_files):
    """Test that report template renders."""
    with current_app.test_client() as test_client:
        response = test_client.get("/reporter/ingest_log_tabular/?amss_id=1")
        assert response.status_code == 200


def test_user_ingest_log_csv(app_with_populated_files):
    """Test CSV export."""
    with current_app.test_client() as test_client:
        response = test_client.get("/reporter/ingest_log_tabular/?amss_id=1&csv=True")
        assert response.status_code == 200
        assert (
            response.headers["Content-Disposition"]
            == "attachment; filename=user_ingest_log.csv"
        )
        assert response.mimetype == "text/csv"
        assert response.data == EXPECTED_CSV_CONTENTS



"""Field name constants."""

FIELD_FILE_FORMAT = "file_format"
FIELD_FILE_TYPE = "file_type"
FIELD_LIMIT = "limit"
FIELD_ORIGINAL_FILES = "original_files"
FIELD_PUID = "puid"
FIELD_STORAGE_LOCATION = "storage_location"
FIELD_START_DATE = "start_date"
FIELD_END_DATE = "end_date"
FIELD_CUMULATIVE = "cumulative"



"""Infos namespace

At the moment, this is just a demonstration endpoint to show how
a namespace can be created and used. We only use it to return version
from the application itself.
"""

from flask_restx import Namespace
from flask_restx import Resource
from flask_restx import fields

import AIPscan

api = Namespace(
    "infos", description="Additional metadata associated with the application"
)

version = api.model(
    "Version",
    {
        "version": fields.String(
            required=True, description="The current version of AIPscan"
        )
    },
)

APPVERSION = {"version": AIPscan.__version__}


@api.route("/version")
class Version(Resource):
    @api.doc("aipscan_version")
    @api.marshal_list_with(version)
    def get(self):
        """Get the application version"""
        return APPVERSION



from flask import current_app


def test_update_dates(app_with_populated_format_versions):
    """Test that report template renders."""
    with current_app.test_client() as test_client:
        get_response = test_client.get("/reporter/update_dates/")
        assert get_response.status_code == 405

        post_response = test_client.post(
            "/reporter/update_dates/", json={"start_date": "2021-02-14"}
        )
        assert post_response.status_code == 200
        assert post_response.json.get("start_date") == "2021-02-14"

        post_response = test_client.post(
            "/reporter/update_dates/", json={"end_date": "2021-02-14"}
        )
        assert post_response.status_code == 200
        assert post_response.json.get("end_date") == "2021-02-14"



import pytest

from AIPscan import db
from AIPscan import typesense_helpers
from AIPscan import typesense_test_helpers
from AIPscan.Data import fields
from AIPscan.Data import report_data
from AIPscan.Data import report_data_typesense
from AIPscan.helpers import parse_datetime_bound
from AIPscan.models import AIP


@pytest.mark.parametrize(
    "storage_location_id, storage_location_description, start_date, end_date, aip_count,largest_aip_size,second_largest_aip_size",
    [
        # Test all AIPs.
        (None, None, "2020-01-01", "2022-12-31", 5, 10000, 750),
        # Test filtering by date.
        (None, None, "2020-01-01", "2020-06-01", 2, 10000, 250),
        # Test filtering by storage location.
        (1, "storage location 1", "2020-01-01", "2022-12-31", 2, 10000, 250),
        (2, "storage location 2", "2020-01-01", "2022-12-31", 3, 750, 500),
    ],
)
def test_largest_aips(
    largest_aips,
    storage_location_id,
    storage_location_description,
    start_date,
    end_date,
    aip_count,
    largest_aip_size,
    second_largest_aip_size,
):
    """Test that outputs of report_data.largest_aips match expectations."""
    report = report_data.largest_aips(
        storage_service_id=1,
        start_date=parse_datetime_bound(start_date),
        end_date=parse_datetime_bound(end_date, upper=True),
        storage_location_id=storage_location_id,
    )
    report_aips = report[fields.FIELD_AIPS]
    assert report[fields.FIELD_STORAGE_NAME] == "test storage service"
    assert report[fields.FIELD_STORAGE_LOCATION] == storage_location_description

    assert len(report_aips) == aip_count
    assert report_aips[0][fields.FIELD_SIZE] == largest_aip_size
    assert report_aips[1][fields.FIELD_SIZE] == second_largest_aip_size


def test_largest_aips_typesense(app_with_populated_files, enable_typesense, mocker):
    doc = typesense_helpers.model_instance_to_document(AIP, db.session.get(AIP, 1))
    fake_results = {"hits": [{"document": doc}]}

    typesense_test_helpers.fake_collection(mocker, fake_results)

    expected_result = {
        "AIPs": [
            {
                "UUID": "111111111111-1111-1111-11111111",
                "Name": "Test AIP",
                "Size": 100,
                "FileCount": 1,
            }
        ],
        "StorageName": "test storage service",
        "StorageLocation": "test storage location",
    }

    report = report_data_typesense.largest_aips(
        storage_service_id=1,
        start_date=parse_datetime_bound("2019-01-01"),
        end_date=parse_datetime_bound("2019-10-10"),
        storage_location_id=1,
    )

    assert report == expected_result



"""This module defines and initalizes a Celery worker.

Since Celery workers are run separately from the Flask application (for
example via a systemd service), we use our Application Factory function
to provide application context.
"""

from AIPscan import create_app
from AIPscan.celery import configure_celery

app = create_app()
celery = configure_celery(app)



"""Initialize components which need to be shared across the report
module, notably too the Blueprint itself.
"""

from flask import Blueprint

from AIPscan.Reporter.helpers import download_csv  # noqa: F401
from AIPscan.Reporter.helpers import format_size_for_csv  # noqa: F401
from AIPscan.Reporter.helpers import get_display_end_date  # noqa: F401
from AIPscan.Reporter.helpers import sort_puids  # noqa: F401
from AIPscan.Reporter.helpers import translate_headers  # noqa: F401

reporter = Blueprint("reporter", __name__, template_folder="templates")



"""Collects a number of functions that aid in the retrieval of
information from an AIP METS file.
"""

import os

import lxml
import metsrw
import requests

from AIPscan.Aggregator.task_helpers import create_numbered_subdirs
from AIPscan.Aggregator.task_helpers import get_mets_url
from AIPscan.helpers import stream_write_and_hash


class METSError(Exception):
    """Exception to signal that we have encountered an error parsing
    the METS document.
    """


def parse_mets_with_metsrw(mets_file):
    """Load and Parse the METS.

    Errors which we encounter at this point will be critical to the
    caller and so an exception is returned when we can't do any better.
    """
    try:
        mets = metsrw.METSDocument.fromfile(mets_file)
    except AttributeError as err:
        # See archivematica/issues#1129 where METSRW expects a certain
        # METS structure but Archivematica has written it incorrectly.
        raise METSError("Error parsing METS: Cannot return a METSDocument") from err
    except lxml.etree.Error as err:
        # We have another undetermined storage service error, e.g. the
        # package no longer exists on the server, or another download
        # error.
        msg = f"Error parsing METS: {err}: {mets_file}"
        raise METSError(msg) from err
    return mets


def get_aip_original_name(mets):
    """Retrieve PREMIS original name from a METSDocument object.

    If the original name cannot be reliably retrieved from the METS file
    a METSError exception is returned to be handled by the caller as
    desired.
    """

    # Negated as we're going to want to remove this length of values.
    NAMESUFFIX = -len("-00000000-0000-0000-0000-000000000000")

    # The transfer directory prefix is a directory prefix that can also
    # exist in a dmdSec intellectual entity and we want to identify and
    # ignore those.
    TRANSFER_DIR_PREFIX = "%transferDirectory%"

    NAMESPACES = {"premis": "http://www.loc.gov/premis/v3"}
    ELEM_ORIGINAL_NAME_PATTERN = ".//premis:originalName"

    original_name = ""
    for fsentry in mets.all_files():
        for dmdsec in fsentry.dmdsecs:
            dmd_element = dmdsec.serialize()
            full_name = dmd_element.find(
                ELEM_ORIGINAL_NAME_PATTERN, namespaces=NAMESPACES
            )
            if full_name is not None and full_name.text.startswith(TRANSFER_DIR_PREFIX):
                # We don't want this value, it will usually represent an
                # directory entity.
                continue
            try:
                original_name = full_name.text[:NAMESUFFIX]
            except AttributeError:
                continue

    # There should be a transfer name in every METS.
    if original_name == "":
        raise METSError("Cannot locate transfer name in METS")

    return original_name


def download_mets(
    storage_service, package_uuid, relative_path_to_mets, timestamp, package_list_no
):
    """Download METS from the storage service.

    The METS document is streamed directly to disk while its SHA256 digest is
    computed so that the caller can avoid re-reading the file.
    """

    numbered_subdir = create_numbered_subdirs(timestamp, package_list_no)
    download_file = os.path.join(numbered_subdir, f"METS.{package_uuid}.xml")

    mets_response = requests.get(
        get_mets_url(storage_service, package_uuid, relative_path_to_mets),
        stream=True,
    )

    try:
        mets_hash = stream_write_and_hash(mets_response, download_file)
    finally:
        mets_response.close()

    return download_file, mets_hash



from operator import itemgetter

from flask import render_template
from flask import request

from AIPscan.Data import data
from AIPscan.Data import fields
from AIPscan.helpers import parse_bool
from AIPscan.helpers import parse_datetime_bound
from AIPscan.Reporter import download_csv
from AIPscan.Reporter import format_size_for_csv
from AIPscan.Reporter import get_display_end_date
from AIPscan.Reporter import reporter
from AIPscan.Reporter import request_params
from AIPscan.Reporter import translate_headers

CSV_HEADERS = [
    fields.FIELD_UUID,
    fields.FIELD_AIP_NAME,
    fields.FIELD_CREATED_DATE,
    fields.FIELD_SIZE,
    fields.FIELD_FORMATS,
]

TABLE_HEADERS = [
    fields.FIELD_AIP_NAME,
    fields.FIELD_CREATED_DATE,
    fields.FIELD_SIZE,
    fields.FIELD_FORMATS,
]


def _create_aip_formats_string_representation(aips, separator="<br>"):
    """Return data prepared for CSV file.

    :param aips: AIPS data returned by data.aip_file_format_overview
        endpoint (list of dicts)

    :returns: Input data with each AIP's formats value formatted as a string
        (list of dicts)
    """
    for aip in aips:
        formats = []
        aip_formats = _create_aip_formats_list_sorted_by_count(
            aip.get(fields.FIELD_FORMATS)
        )
        for format_ in aip_formats:
            plural = ""
            count = format_.get(fields.FIELD_COUNT)
            if count > 1:
                plural = "s"
            format_string = f"{format_.get(fields.FIELD_PUID)} ({format_.get(fields.FIELD_FORMAT)}): {count} file{plural}"
            formats.append(format_string)
        aip[fields.FIELD_FORMATS] = f"{separator}".join(
            [format_ for format_ in formats]
        )
    return aips


def _create_aip_formats_list_sorted_by_count(formats):
    """Return list of AIP format dicts sorted by count from input dict.

    The dict returned for each AIP in fields.FIELD_FORMATS by the data module's
    aip_file_format_overview endpoint is difficult to sort, as it is a dict of
    dicts with an unpredictable key (PUID or format name) for each child
    dict. This function transforms this into an easier to parse list of dicts,
    sorted descending by count.

    :param formats: Dict of dicts returned by data.aip_file_format_overview for
        each aip (dict)

    :returns: Input data transformed into a sorted list of dicts.
    """
    formats_list = []
    for key, values in formats.items():
        format_info = {}
        format_info[fields.FIELD_PUID] = key
        format_info[fields.FIELD_FORMAT] = values.get(fields.FIELD_NAME)
        if values.get(fields.FIELD_VERSION):
            format_info[fields.FIELD_FORMAT] = (
                f"{values.get(fields.FIELD_NAME)} {values.get(fields.FIELD_VERSION)}"
            )
        format_info[fields.FIELD_COUNT] = values.get(fields.FIELD_COUNT)
        formats_list.append(format_info)
    return sorted(formats_list, key=itemgetter(fields.FIELD_COUNT), reverse=True)


@reporter.route("/aip_contents/", methods=["GET"])
def aip_contents():
    """Return AIP contents organized by format."""
    storage_service_id = request.args.get(request_params.STORAGE_SERVICE_ID)
    storage_location_id = request.args.get(request_params.STORAGE_LOCATION_ID)
    start_date = parse_datetime_bound(request.args.get(request_params.START_DATE))
    end_date = parse_datetime_bound(
        request.args.get(request_params.END_DATE), upper=True
    )
    csv = parse_bool(request.args.get(request_params.CSV), default=False)

    aip_data = data.aip_file_format_overview(
        storage_service_id=storage_service_id,
        start_date=start_date,
        end_date=end_date,
        storage_location_id=storage_location_id,
    )

    if csv:
        headers = translate_headers(CSV_HEADERS, True)

        filename = "aip_contents.csv"
        aips = _create_aip_formats_string_representation(
            aip_data.get(fields.FIELD_AIPS), separator="|"
        )
        csv_data = format_size_for_csv(aips)
        return download_csv(headers, csv_data, filename)

    aips = _create_aip_formats_string_representation(aip_data.get(fields.FIELD_AIPS))

    headers = translate_headers(TABLE_HEADERS)

    return render_template(
        "report_aip_contents.html",
        storage_service=storage_service_id,
        storage_service_name=aip_data.get(fields.FIELD_STORAGE_NAME),
        storage_location_description=aip_data.get(fields.FIELD_STORAGE_LOCATION),
        columns=headers,
        aips=aip_data.get(fields.FIELD_AIPS),
        start_date=start_date,
        end_date=get_display_end_date(end_date),
    )



import os

import pytest
from flask import current_app

SCRIPT_DIR = os.path.dirname(os.path.realpath(__file__))
FIXTURES_DIR = os.path.join(SCRIPT_DIR, "fixtures")

EXPECTED_CSV_CONTENTS = b"UUID,Location,AIPs,Size,Size (bytes),File Count\r\n2bbcea40-eb4d-4076-a81d-1ab046e34f6a,AIP Store Location 1,2,1.6 kB,1600,3\r\ne69beb57-0e32-4c45-8db7-9b7723724a05,AIP Store Location 2,1,5.0 kB,5000,2\r\n"


def test_storage_locations(storage_locations):
    """Test that report template renders."""
    with current_app.test_client() as test_client:
        response = test_client.get("/reporter/storage_locations/?amss_id=1")
        assert response.status_code == 200


def test_storage_locations_csv(storage_locations):
    """Test CSV export."""
    with current_app.test_client() as test_client:
        response = test_client.get("/reporter/storage_locations/?amss_id=1&csv=True")
        assert response.status_code == 200
        assert (
            response.headers["Content-Disposition"]
            == "attachment; filename=storage_locations.csv"
        )
        assert response.mimetype == "text/csv"
        assert response.data == EXPECTED_CSV_CONTENTS


def test_storage_locations_usage_over_time(storage_locations):
    """Test that report template renders."""
    with current_app.test_client() as test_client:
        response = test_client.get(
            "/reporter/storage_locations_usage_over_time/?amss_id=1"
        )
        assert response.status_code == 200


@pytest.mark.parametrize(
    "metric, cumulative, fixture",
    [
        # Differential totals
        ("aips", False, "storage_location_usage_differential_aips.csv"),
        ("size", False, "storage_location_usage_differential_size.csv"),
        ("files", False, "storage_location_usage_differential_files.csv"),
        # Cumulative totals
        ("aips", True, "storage_location_usage_cumulative_aips.csv"),
        ("size", True, "storage_location_usage_cumulative_size.csv"),
        ("files", True, "storage_location_usage_cumulative_files.csv"),
    ],
)
def test_storage_locations_usage_over_time_csv(
    storage_locations, metric, cumulative, fixture
):
    """Test CSV export."""
    with current_app.test_client() as test_client:
        response = test_client.get(
            f"/reporter/storage_locations_usage_over_time/?amss_id=1&csv=True&start_date=2020-01-01&end_date=2020-06-30&metric={metric}&cumulative={cumulative}"
        )
        assert response.status_code == 200
        assert (
            response.headers["Content-Disposition"]
            == "attachment; filename=storage_locations_usage_over_time.csv"
        )
        assert response.mimetype == "text/csv"

        csv_fixture_path = os.path.join(FIXTURES_DIR, fixture)
        with open(csv_fixture_path, "rb") as csv_file:
            assert response.data == csv_file.read()



import json
import os
from datetime import datetime

import pytest

from AIPscan import models
from AIPscan.Aggregator import task_helpers
from AIPscan.Aggregator.types import StorageServicePackage

FIXTURES_DIR = "fixtures"

LOCATION_UUID = "1b60c346-85a0-4a3c-a88b-0c1b3255e2ec"


@pytest.mark.parametrize(
    "input_date,output_date,now_year",
    [
        ("2020-05-19T08:04:16+00:00", "2020-05-19T08:04:16", False),
        ("2020-07-30T13:27:45.757482+00:00", "2020-07-30T13:27:45", False),
        ("2020-07-30", "2020-07-30T00:00:00", False),
        ("T13:27:45", "", True),
        ("こんにちは世界", "1970-01-01T00:00:01", False),
    ],
)
def test_tz_neutral_dates(input_date, output_date, now_year):
    """Ensure datetime values are handled sensibly across regions."""
    result_date = task_helpers._tz_neutral_date(input_date)
    if now_year is True:
        year = datetime.now().strftime("%Y-%m-%d")
        output_date = f"{year}{input_date}"
        output_date = datetime.strptime(output_date, "%Y-%m-%dT%H:%M:%S")
        assert result_date == output_date
    else:
        output_date = datetime.strptime(output_date, "%Y-%m-%dT%H:%M:%S")
        assert result_date == output_date


@pytest.mark.parametrize(
    "ss_args, base_url, url_without_api_key, url_with_api_key",
    [
        (
            {
                "name": "Test",
                "url": "http://example.com:9000/",
                "user_name": "test",
                "api_key": "mykey",
                "download_limit": "23",
                "download_offset": "13",
                "default": False,
            },
            "http://example.com:9000",
            "http://example.com:9000/api/v2/file/?limit=23&offset=13",
            "http://example.com:9000/api/v2/file/?limit=23&offset=13&username=test&api_key=mykey",
        ),
        (
            {
                "name": "Test",
                "url": "http://subdomain.example.com:8000/",
                "user_name": "anothertest",
                "api_key": "myotherkey",
                "download_limit": "10",
                "download_offset": "99",
                "default": False,
            },
            "http://subdomain.example.com:8000",
            "http://subdomain.example.com:8000/api/v2/file/?limit=10&offset=99",
            "http://subdomain.example.com:8000/api/v2/file/?limit=10&offset=99&username=anothertest&api_key=myotherkey",
        ),
    ],
)
def test_format_api_url(ss_args, base_url, url_without_api_key, url_with_api_key):
    storage_service = models.StorageService(**ss_args)
    res1, res2, res3 = task_helpers.format_api_url_with_limit_offset(storage_service)
    assert res1 == base_url
    assert res2 == url_without_api_key
    assert res3 == url_with_api_key


@pytest.mark.parametrize(
    "ss_args, package_uuid, path_to_mets, result",
    [
        (
            {
                "name": "Test",
                "url": "http://example.com",
                "user_name": "1234",
                "api_key": "1234",
                "download_limit": 0,
                "download_offset": 0,
                "default": False,
            },
            "1234",
            "1234",
            "http://example.com/api/v2/file/1234/extract_file/?relative_path_to_file=1234&username=1234&api_key=1234",
        ),
        (
            {
                "name": "Test",
                "url": "http://example.com",
                "user_name": "1234",
                "api_key": "1234",
                "download_limit": 0,
                "download_offset": 0,
                "default": False,
            },
            "1234",
            "1234",
            "http://example.com/api/v2/file/1234/extract_file/?relative_path_to_file=1234&username=1234&api_key=1234",
        ),
    ],
)
def test_get_mets_url(ss_args, package_uuid, path_to_mets, result):
    """Ensure that the URL for retrieving METS is constructed properly."""
    ss = models.StorageService(**ss_args)
    mets_url = task_helpers.get_mets_url(ss, package_uuid, path_to_mets)
    assert mets_url == result


@pytest.mark.parametrize(
    "ss_args, current_location, expected_url, expected_url_without_api_key",
    [
        (
            {
                "name": "Test",
                "url": "http://example.com",
                "user_name": "1234",
                "api_key": "12345",
                "download_limit": 0,
                "download_offset": 0,
                "default": False,
            },
            f"/api/v2/location/{LOCATION_UUID}",
            "http://example.com/api/v2/location/1b60c346-85a0-4a3c-a88b-0c1b3255e2ec?username=1234&api_key=12345",
            "http://example.com/api/v2/location/1b60c346-85a0-4a3c-a88b-0c1b3255e2ec",
        )
    ],
)
def test_get_storage_service_api_url(
    ss_args, current_location, expected_url, expected_url_without_api_key
):
    """Ensure construction of URL to fetch Resource information."""
    storage_service = models.StorageService(**ss_args)

    url, url_without_secrets = task_helpers.get_storage_service_api_url(
        storage_service, current_location
    )
    assert url == expected_url
    assert url_without_secrets == expected_url_without_api_key


@pytest.mark.parametrize("timestamp, package_list_number", [("1234", 1)])
def test_create_numbered_subdirs(timestamp, package_list_number, mocker, tmp_path):
    """Ensure that the logic that we use to create sub-directories for
    storing METS is sound.
    """

    download_root = tmp_path / "aipscan-downloads"
    expected_path = os.path.join(
        os.fspath(download_root), timestamp, "mets", str(package_list_number)
    )

    mocker.patch(
        "AIPscan.Aggregator.task_helpers.get_download_root",
        return_value=os.fspath(download_root),
    )

    # Test that an unknown directory is created first time around and
    # that the correct result is returned.
    mocker.patch("os.path.exists", return_value=False)
    mocked_makedirs = mocker.patch("os.makedirs")
    subdir_string = task_helpers.create_numbered_subdirs(timestamp, package_list_number)
    mocked_makedirs.assert_called_once_with(expected_path)
    assert subdir_string == expected_path

    # Test that if the path exists, we don't create the directory, and
    # that the correct result is returned.
    mocker.patch("os.path.exists", return_value=True)
    mocked_makedirs = mocker.patch("os.makedirs")
    subdir_string = task_helpers.create_numbered_subdirs(timestamp, package_list_number)
    assert mocked_makedirs.call_count == 0
    assert subdir_string == expected_path


@pytest.fixture()
def packages():
    fixtures_path = os.path.join("package_json", "packages.json")
    script_dir = os.path.dirname(os.path.realpath(__file__))
    packages_file = os.path.join(script_dir, FIXTURES_DIR, fixtures_path)
    with open(packages_file) as package_json:
        packages = json.load(package_json)
        return packages.get("objects")


@pytest.mark.parametrize(
    "idx, storage_service_package",
    [
        (
            0,
            StorageServicePackage(
                aip=True,
                current_path="9194/0daf/5c33/4670/bfc8/9108/f32b/ca7b/repl-91940daf-5c33-4670-bfc8-9108f32bca7b.7z",
                uuid="91940daf-5c33-4670-bfc8-9108f32bca7b",
                current_location="/api/v2/location/a083fd8c-8b5a-4ece-acc4-c6388ade3d56/",
                origin_pipeline="/api/v2/pipeline/dbb92514-1572-4a81-87ee-06408672e7a6/",
            ),
        ),
        (
            1,
            StorageServicePackage(
                replica=True,
                aip=True,
                current_path="9021/92d9/6232/4d5f/b6c4/dfa7/b873/3960/repl-902192d9-6232-4d5f-b6c4-dfa7b8733960.7z",
                uuid="902192d9-6232-4d5f-b6c4-dfa7b8733960",
                current_location="/api/v2/location/a82c7f40-34f5-4442-a408-cbcf16f0c1cf/",
                origin_pipeline="/api/v2/pipeline/dbb92514-1572-4a81-87ee-06408672e7a6/",
            ),
        ),
        (
            2,
            StorageServicePackage(
                deleted=True,
                aip=True,
                current_path="594a/03f3/d8eb/4c83/affd/aefa/f75d/53cc/delete_me-594a03f3-d8eb-4c83-affd-aefaf75d53cc.7z",
                uuid="594a03f3-d8eb-4c83-affd-aefaf75d53cc",
                current_location="/api/v2/location/a083fd8c-8b5a-4ece-acc4-c6388ade3d56/",
                origin_pipeline="/api/v2/pipeline/dbb92514-1572-4a81-87ee-06408672e7a6/",
            ),
        ),
        (
            3,
            StorageServicePackage(
                aip=True,
                current_path="e422/c724/834c/4164/a7be/4c88/1043/a531/normalize-e422c724-834c-4164-a7be-4c881043a531.7z",
                uuid="e422c724-834c-4164-a7be-4c881043a531",
                current_location="/api/v2/location/a083fd8c-8b5a-4ece-acc4-c6388ade3d56/",
                origin_pipeline="/api/v2/pipeline/dbb92514-1572-4a81-87ee-06408672e7a6/",
            ),
        ),
        (
            4,
            StorageServicePackage(
                aip=True,
                current_path="583c/009c/4255/44b9/8868/f57e/4bad/e93c/normal_aip-583c009c-4255-44b9-8868-f57e4bade93c.7z",
                uuid="583c009c-4255-44b9-8868-f57e4bade93c",
                current_location="/api/v2/location/a083fd8c-8b5a-4ece-acc4-c6388ade3d56/",
                origin_pipeline="/api/v2/pipeline/dbb92514-1572-4a81-87ee-06408672e7a6/",
            ),
        ),
        (
            5,
            StorageServicePackage(
                aip=True,
                current_path="846f/ca2b/0919/4673/804c/0f62/6a30/cabd/uncomp-846fca2b-0919-4673-804c-0f626a30cabd",
                uuid="846fca2b-0919-4673-804c-0f626a30cabd",
                current_location="/api/v2/location/a083fd8c-8b5a-4ece-acc4-c6388ade3d56/",
                origin_pipeline="/api/v2/pipeline/dbb92514-1572-4a81-87ee-06408672e7a6/",
            ),
        ),
        (
            6,
            StorageServicePackage(
                dip=True,
                current_path="1555/04f5/98dd/48d1/9e97/83bc/f0a5/efa2/normcore-59f70134-eeca-4886-888e-b2013a08571e",
                uuid="155504f5-98dd-48d1-9e97-83bcf0a5efa2",
                current_location="/api/v2/location/7e99c7b0-bd30-4d49-b173-3cb59a32f0c5/",
                origin_pipeline="/api/v2/pipeline/dbb92514-1572-4a81-87ee-06408672e7a6/",
            ),
        ),
        (
            7,
            StorageServicePackage(
                aip=True,
                current_path="59f7/0134/eeca/4886/888e/b201/3a08/571e/normcore-59f70134-eeca-4886-888e-b2013a08571e",
                uuid="59f70134-eeca-4886-888e-b2013a08571e",
                current_location="/api/v2/location/a083fd8c-8b5a-4ece-acc4-c6388ade3d56/",
                origin_pipeline="/api/v2/pipeline/dbb92514-1572-4a81-87ee-06408672e7a6/",
            ),
        ),
        (
            8,
            StorageServicePackage(
                sip=True,
                current_path="originals/backlog-fbdcd607-270e-4dff-9a01-d11b7c2a0200",
                uuid="fbdcd607-270e-4dff-9a01-d11b7c2a0200",
                current_location="/api/v2/location/7fe31789-90e0-4ad3-a669-18517737fc25/",
                origin_pipeline="/api/v2/pipeline/dbb92514-1572-4a81-87ee-06408672e7a6/",
            ),
        ),
    ],
)
def test_process_package_object(packages, idx, storage_service_package):
    """Test our ability to collect information from a package list
    object and parse it into a storage service package type for
    further use.
    """
    package_obj = task_helpers.process_package_object(packages[idx])
    assert package_obj == storage_service_package, idx


def test_summarize_fetch_job_results():
    fetch_job = models.FetchJob(
        total_packages=15,
        total_aips=1,
        total_deleted_aips=4,
        download_start=None,
        download_end=None,
        download_directory=None,
        storage_service_id=None,
    )
    fetch_job.total_sips = 2
    fetch_job.total_dips = 3
    fetch_job.total_replicas = 5

    assert (
        "aips: '1'; sips: '2'; dips: '3'; deleted: '4'; replicated: '5'"
        == task_helpers.summarize_fetch_job_results(fetch_job)
    )



import pytest
from flask import current_app

from AIPscan import test_helpers
from AIPscan.Aggregator.tasks import TaskError
from AIPscan.Aggregator.views import _test_storage_service_connection
from AIPscan.models import StorageService


def test_new_fetch_job_success_returns_task_id(app_instance, mocker):
    ss = test_helpers.create_test_storage_service()

    mocker.patch("AIPscan.Aggregator.views._test_storage_service_connection")
    mocker.patch("AIPscan.Aggregator.views.os.makedirs")

    class _FakeTask:
        id = "coordinator-123"

    mocker.patch(
        "AIPscan.Aggregator.views.tasks.workflow_coordinator.delay",
        return_value=_FakeTask(),
    )

    mocker.patch(
        "AIPscan.Aggregator.views._wait_for_package_list_task_id",
        return_value="child-456",
    )

    with current_app.test_client() as test_client:
        response = test_client.post(f"/aggregator/new_fetch_job/{ss.id}")
        assert response.status_code == 200

        data = response.get_json()
        assert set(["timestamp", "taskId", "fetchJobId"]).issubset(data.keys())
        assert data["taskId"] == "child-456"
        assert isinstance(data["fetchJobId"], int)


def test_new_fetch_job_timeout_returns_500(app_instance, mocker):
    """If no child id appears, the view responds with 500 JSON error."""
    ss = test_helpers.create_test_storage_service()

    mocker.patch("AIPscan.Aggregator.views._test_storage_service_connection")
    mocker.patch("AIPscan.Aggregator.views.os.makedirs")

    class _FakeTask:
        id = "coordinator-xyz"

    mocker.patch(
        "AIPscan.Aggregator.views.tasks.workflow_coordinator.delay",
        return_value=_FakeTask(),
    )

    # Simulate helper timing out.
    mocker.patch(
        "AIPscan.Aggregator.views._wait_for_package_list_task_id",
        return_value=None,
    )

    with current_app.test_client() as test_client:
        response = test_client.post(f"/aggregator/new_fetch_job/{ss.id}")
        assert response.status_code == 500

        data = response.get_json()
        assert "message" in data
        assert "failed to start" in data["message"].lower()


def test__test_storage_service_connection(mocker):
    """Test that invalid SS connection raises ConnectionError."""
    make_request = mocker.patch("AIPscan.Aggregator.views.tasks.make_request")
    make_request.side_effect = TaskError("Bad response from server")

    with pytest.raises(ConnectionError):
        storage_service = StorageService(
            "Test", "http://test-url", "test", "test", "0", "10", ""
        )
        _test_storage_service_connection(storage_service)


def test_new_fetch_job_bad_connection(app_with_populated_files, mocker):
    """Verify Celery tasks don't run wihtout a Storage Service connection."""
    task = mocker.patch("AIPscan.Aggregator.views.tasks.workflow_coordinator.delay")
    test_storage_service_conn = mocker.patch(
        "AIPscan.Aggregator.views._test_storage_service_connection"
    )
    test_storage_service_conn.side_effect = ConnectionError("Bad response from server")
    with current_app.test_client() as test_client:
        response = test_client.post("/aggregator/new_fetch_job/1")
        task.assert_not_called()
        assert response.status_code == 400


def test_view_storage_service(app_with_populated_files):
    with current_app.test_client() as test_client:
        response = test_client.get("/aggregator/storage_service/1")
        assert response.status_code == 200

        response = test_client.get("/aggregator/storage_service/0")
        assert response.status_code == 404


def test_edit_storage_service(app_with_populated_files):
    with current_app.test_client() as test_client:
        response = test_client.post("/aggregator/edit_storage_service/1")
        assert response.status_code == 200

        response = test_client.post("/aggregator/edit_storage_service/0")
        assert response.status_code == 404


def test_delete_storage_service(app_with_populated_files, mocker):
    with current_app.test_client() as test_client:
        # The delete confirmation page should not show up for a non-existent storage service
        response = test_client.get("/aggregator/delete_storage_service/0")
        assert response.status_code == 404

        # Deletion should not be attempted for a non-existent storage service
        response = test_client.get("/aggregator/delete_storage_service/0?confirm=1")
        assert response.status_code == 404

        # Don't actually instantiate Celery job
        mocker.patch("AIPscan.Aggregator.tasks.delete_storage_service.delay")

        # The delete confirmation page should show up for an existing storage service
        response = test_client.get("/aggregator/delete_storage_service/1")
        assert response.status_code == 200

        # Deletion should be attempted if the storage service exists
        response = test_client.get("/aggregator/delete_storage_service/1?confirm=1")
        assert response.status_code == 200


def test_delete_fetch_job(app_with_populated_files, mocker):
    with current_app.test_client() as test_client:
        response = test_client.get("/aggregator/delete_fetch_job/0?confirm=1")
        assert response.status_code == 404

        mocker.patch("AIPscan.Aggregator.tasks.delete_fetch_job.delay")

        response = test_client.get("/aggregator/delete_fetch_job/1?confirm=1")
        assert response.status_code == 302



from datetime import datetime

import pytest

from AIPscan.conftest import AIP_1_CREATION_DATE
from AIPscan.conftest import AIP_2_CREATION_DATE
from AIPscan.conftest import ORIGINAL_FILE_SIZE as JPEG_1_01_FILE_SIZE
from AIPscan.conftest import PRESERVATION_FILE_SIZE as JPEG_1_02_FILE_SIZE
from AIPscan.Data import fields
from AIPscan.Data import report_data
from AIPscan.Data import report_data_typesense
from AIPscan.Data.tests import MOCK_STORAGE_SERVICE
from AIPscan.Data.tests import MOCK_STORAGE_SERVICE_ID
from AIPscan.Data.tests import MOCK_STORAGE_SERVICE_NAME
from AIPscan.helpers import parse_datetime_bound
from AIPscan.test_helpers import create_test_storage_location

TOTAL_FILE_SIZE = JPEG_1_01_FILE_SIZE + JPEG_1_02_FILE_SIZE

DATE_BEFORE_AIP_1 = "2019-01-01"
DATE_AFTER_AIP_1 = "2020-01-02"
DATE_BEFORE_AIP_2 = "2020-05-30"
DATE_AFTER_AIP_2 = "2020-06-02"


class MockFormatVersionsCountQueryResult:
    """Fixture for mocking SQLAlchemy query results."""

    def __init__(self, puid, file_format, format_version, file_count, total_size):
        self.puid = puid
        self.file_format = file_format
        self.format_version = format_version
        self.file_count = file_count
        self.total_size = total_size


MOCK_QUERY_RESULTS = [
    MockFormatVersionsCountQueryResult(
        puid="fmt/43",
        file_format="JPEG",
        format_version="1.01",
        file_count=5,
        total_size=12345678,
    ),
    MockFormatVersionsCountQueryResult(
        puid="fmt/44",
        file_format="JPEG",
        format_version="1.02",
        file_count=3,
        total_size=1234567,
    ),
    MockFormatVersionsCountQueryResult(
        puid="fmt/199",
        file_format="MPEG-4 Media File",
        format_version=None,
        file_count=1,
        total_size=12345,
    ),
]


@pytest.mark.parametrize(
    "query_results, results_count",
    [
        # Empty result set, count is 0.
        ([], 0),
        # Test the return of complete result set, count is the length
        # of all results.
        (MOCK_QUERY_RESULTS, len(MOCK_QUERY_RESULTS)),
        # Test the return of only the first two results, count is 2.
        (MOCK_QUERY_RESULTS[:2], 2),
    ],
)
def test_format_versions_count(app_instance, mocker, query_results, results_count):
    """Test that results match high-level expectations."""
    query = mocker.patch("AIPscan.Data.report_data._format_versions_count_query")
    query.return_value = query_results

    get_ss = mocker.patch("AIPscan.Data._get_storage_service")
    get_ss.return_value = MOCK_STORAGE_SERVICE

    test_location = create_test_storage_location()
    get_location = mocker.patch("AIPscan.Data._get_storage_location")
    get_location.return_value = test_location

    report = report_data.format_versions_count(
        storage_service_id=MOCK_STORAGE_SERVICE_ID,
        start_date=datetime.min,
        end_date=datetime.max,
        storage_location_id=test_location.id,
    )
    assert report[fields.FIELD_STORAGE_NAME] == MOCK_STORAGE_SERVICE_NAME
    assert report[fields.FIELD_STORAGE_LOCATION] == test_location.description
    assert len(report[fields.FIELD_FORMAT_VERSIONS]) == results_count


@pytest.mark.parametrize(
    "test_format_version", [mock_result for mock_result in MOCK_QUERY_RESULTS]
)
def test_format_versions_count_elements(app_instance, mocker, test_format_version):
    """Test that structure of versions data matches expectations."""
    mock_query = mocker.patch("AIPscan.Data.report_data._format_versions_count_query")
    mock_query.return_value = [test_format_version]

    mock_get_ss_name = mocker.patch("AIPscan.Data._get_storage_service")
    mock_get_ss_name.return_value = MOCK_STORAGE_SERVICE

    report = report_data.format_versions_count(
        MOCK_STORAGE_SERVICE_ID, datetime.min, datetime.max
    )
    report_format_version = report[fields.FIELD_FORMAT_VERSIONS][0]

    assert test_format_version.puid == report_format_version.get(fields.FIELD_PUID)
    assert test_format_version.file_format == report_format_version.get(
        fields.FIELD_FORMAT
    )
    assert test_format_version.format_version == report_format_version.get(
        fields.FIELD_VERSION
    )
    assert test_format_version.file_count == report_format_version.get(
        fields.FIELD_COUNT
    )
    assert test_format_version.total_size == report_format_version.get(
        fields.FIELD_SIZE
    )


@pytest.mark.parametrize(
    "start_date, end_date, version_count, total_file_count, total_file_size",
    [
        # Not specifying dates should return all files and versions.
        (None, None, 3, 3, TOTAL_FILE_SIZE),
        # Start date before first AIP was ingested hould return all
        # files and versions.
        (DATE_BEFORE_AIP_1, None, 3, 3, TOTAL_FILE_SIZE),
        # Start date that's the same day our first AIP was ingested
        # should return all files and versions.
        (AIP_1_CREATION_DATE, None, 3, 3, TOTAL_FILE_SIZE),
        # Start date after our first AIP was ingested should return
        # only the second JPEG version and ISO disk image.
        (DATE_AFTER_AIP_1, None, 2, 2, JPEG_1_02_FILE_SIZE),
        # End date before second AIP was ingested should return only
        # the first JPEG version.
        (None, DATE_BEFORE_AIP_2, 1, 1, JPEG_1_01_FILE_SIZE),
        # End date that's the same day our second AIP was ingested
        # should return all files and versions.
        (None, AIP_2_CREATION_DATE, 3, 3, TOTAL_FILE_SIZE),
        # End date that's after our second AIP was ingested should
        # return all files and versions.
        (None, DATE_AFTER_AIP_2, 3, 3, TOTAL_FILE_SIZE),
        # Start and end dates that define a range in which we haven't
        # ingested any AIPs should return no files or versions.
        ("2019-01-01", "2019-01-02", 0, 0, 0),
        # Invalid values for start and end dates should be treated as
        # None values and return both JPEG versions.
        (True, "NOT A DATE", 3, 3, TOTAL_FILE_SIZE),
    ],
)
def test_format_versions_count_contents(
    app_with_populated_format_versions,
    start_date,
    end_date,
    version_count,
    total_file_count,
    total_file_size,
):
    """Test that content of response matches expectations.

    This integration test uses a pre-populated fixture to verify that
    the database access layer of our endpoint returns what we expect.
    """
    results = report_data.format_versions_count(
        storage_service_id=1,
        start_date=parse_datetime_bound(start_date),
        end_date=parse_datetime_bound(end_date, upper=True),
    )
    versions = results[fields.FIELD_FORMAT_VERSIONS]
    assert len(versions) == version_count
    assert (
        sum(version.get(fields.FIELD_COUNT, 0) for version in versions)
        == total_file_count
    )
    assert (
        sum(version.get(fields.FIELD_SIZE, 0) for version in versions)
        == total_file_size
    )


def test_format_versions_count_no_results_typesense(
    app_with_populated_files, enable_typesense, mocker
):
    mocker.patch("typesense.collections.Collections.__getitem__")
    mocker.patch("typesense.multi_search.MultiSearch.perform")
    mocker.patch("AIPscan.typesense_helpers.facet_value_counts")

    expected_result = {
        "StorageName": "test storage service",
        "StorageLocation": "test storage location",
        "FormatVersions": [],
    }

    start_date = "2019-01-01"
    end_date = "2019-10-01"

    report = report_data_typesense.format_versions_count(
        1,
        datetime.strptime(start_date, "%Y-%m-%d"),
        datetime.strptime(end_date, "%Y-%m-%d"),
        1,
    )
    assert report == expected_result



from flask import current_app

EXPECTED_CSV_CONTENTS = b"AIP Name,UUID,AIP Size,AIP Size (bytes),File Count\r\nTest AIP,111111111111-1111-1111-11111111,100 Bytes,100,1\r\nTest AIP,222222222222-2222-2222-22222222,100 Bytes,100,2\r\n"


def test_largest_aips(app_with_populated_format_versions):
    """Test that report template renders."""
    with current_app.test_client() as test_client:
        response = test_client.get("/reporter/largest_aips/?amss_id=1")
        assert response.status_code == 200


def test_largest_aips_csv(app_with_populated_format_versions):
    """Test CSV export."""
    with current_app.test_client() as test_client:
        response = test_client.get("/reporter/largest_aips/?amss_id=1&csv=True")
        assert response.status_code == 200
        assert (
            response.headers["Content-Disposition"]
            == "attachment; filename=largest_aips.csv"
        )
        assert response.mimetype == "text/csv"
        assert response.data == EXPECTED_CSV_CONTENTS



from flask import render_template
from flask import request

from AIPscan.Data import fields
from AIPscan.Data import report_data
from AIPscan.helpers import parse_bool
from AIPscan.models import File
from AIPscan.Reporter import download_csv
from AIPscan.Reporter import format_size_for_csv
from AIPscan.Reporter import reporter
from AIPscan.Reporter import request_params
from AIPscan.Reporter import translate_headers

HEADERS = [
    fields.FIELD_AIP_NAME,
    fields.FIELD_UUID,
    fields.FIELD_COUNT,
    fields.FIELD_SIZE,
]


def get_format_string_from_puid(puid):
    """Return file format and version info for a PUID or None

    :param puid: PUID (str)

    :returns: "File format (version)", "File Format", or None.
    """
    file_with_puid = File.query.filter_by(puid=puid).first()
    if file_with_puid is None:
        return None
    try:
        file_format = file_with_puid.file_format
    except AttributeError:
        return None
    try:
        format_version = file_with_puid.format_version
    except AttributeError:
        pass
    if format_version:
        return f"{file_format} ({format_version})"
    return file_format


@reporter.route("/aips_by_puid/", methods=["GET"])
def aips_by_puid():
    """Return AIPs containing PUID, sorted by count and total size."""
    storage_service_id = request.args.get(request_params.STORAGE_SERVICE_ID)
    storage_location_id = request.args.get(request_params.STORAGE_LOCATION_ID)
    puid = request.args.get(request_params.PUID)
    original_files = parse_bool(request.args.get(request_params.ORIGINAL_FILES, True))
    csv = parse_bool(request.args.get(request_params.CSV), default=False)

    aip_data = report_data.aips_by_puid(
        storage_service_id=storage_service_id,
        puid=puid,
        original_files=original_files,
        storage_location_id=storage_location_id,
    )

    if csv:
        headers = translate_headers(HEADERS, True)

        filename = f"aips_by_puid_{puid}.csv"
        csv_data = format_size_for_csv(aip_data[fields.FIELD_AIPS])
        return download_csv(headers, csv_data, filename)

    headers = translate_headers(HEADERS)

    return render_template(
        "report_aips_by_puid.html",
        storage_service_id=storage_service_id,
        storage_service_name=aip_data.get(fields.FIELD_STORAGE_NAME),
        storage_location_description=aip_data.get(fields.FIELD_STORAGE_LOCATION),
        puid=puid,
        file_format=get_format_string_from_puid(puid),
        original_files=original_files,
        columns=translate_headers(headers),
        aips=aip_data.get(fields.FIELD_AIPS),
    )



import json

RESPONSE_DICT = {"key": "value"}
VALID_JSON = json.dumps(RESPONSE_DICT)
INVALID_JSON = "test"

REQUEST_URL_WITHOUT_API_KEY = "http://test-archivematica.example.com:8000"
REQUEST_URL = REQUEST_URL_WITHOUT_API_KEY + "?user=test&api_key=test"


class MockResponse:
    """Mock requests.Response class."""

    def __init__(self, status_code, data):
        self.status_code = status_code
        self.data = data

    def json(self):
        return json.loads(self.data)



"""Data endpoints optimized for providing general overviews of AIPs."""

from AIPscan.Data import fields
from AIPscan.Data import report_dict
from AIPscan.helpers import _simplify_datetime
from AIPscan.models import AIP
from AIPscan.models import File
from AIPscan.models import FileType
from AIPscan.models import StorageService


def storage_services():
    """Return a summary overview of storage services."""
    report = {}

    storage_services = []
    for storage_service in StorageService.query.all():
        storage_services.append(
            {"id": storage_service.id, "name": storage_service.name}
        )

    report["storage_services"] = storage_services
    return report


def file_format_aip_overview(
    storage_service_id, original_files=True, storage_location_id=None
):
    """Return summary overview of file formats and the AIPs they're in."""
    report = report_dict(storage_service_id, storage_location_id)
    formats = {}
    aips = AIP.query.filter_by(storage_service_id=storage_service_id)
    if storage_location_id:
        aips = aips.filter_by(storage_location_id=storage_location_id)
    aips = aips.all()
    for aip in aips:
        files = None
        if original_files is True:
            files = File.query.filter_by(aip_id=aip.id, file_type=FileType.original)
        else:
            files = File.query.filter_by(aip_id=aip.id, file_type=FileType.preservation)
        for file_ in files:
            # Originals have PUIDs but Preservation Masters don't.
            # Return a key (PUID or Format Name) for our report based on that.
            try:
                format_key = file_.puid
            except AttributeError:
                format_key = file_.file_format
            if format_key in report:
                formats[format_key][fields.FIELD_COUNT] = (
                    formats[format_key][fields.FIELD_COUNT] + 1
                )
                if aip.uuid not in formats[format_key][fields.FIELD_AIPS]:
                    formats[format_key][fields.FIELD_AIPS].append(aip.uuid)
            else:
                formats[format_key] = {}
                formats[format_key][fields.FIELD_COUNT] = 1
                try:
                    formats[format_key][fields.FIELD_VERSION] = file_.format_version
                    formats[format_key][fields.FIELD_NAME] = file_.file_format
                except AttributeError:
                    pass
                if formats[format_key].get(fields.FIELD_AIPS) is None:
                    formats[format_key][fields.FIELD_AIPS] = []
                formats[format_key][fields.FIELD_AIPS].append(aip.uuid)

    report[fields.FIELD_FORMATS] = formats

    return report


def aip_file_format_overview(
    storage_service_id,
    start_date,
    end_date,
    original_files=True,
    storage_location_id=None,
):
    """Return summary overview of AIPs and their file formats."""
    report = report_dict(storage_service_id, storage_location_id)
    report[fields.FIELD_AIPS] = []
    formats = {}

    aips = AIP.query.filter_by(storage_service_id=storage_service_id)
    if storage_location_id:
        aips = aips.filter_by(storage_location_id=storage_location_id)
    aips = aips.all()

    for aip in aips:
        if aip.create_date < start_date or aip.create_date >= end_date:
            continue

        aip_info = {}
        aip_info[fields.FIELD_UUID] = aip.uuid
        aip_info[fields.FIELD_AIP_NAME] = aip.transfer_name
        aip_info[fields.FIELD_CREATED_DATE] = _simplify_datetime(aip.create_date, False)
        aip_info[fields.FIELD_SIZE] = 0
        aip_info[fields.FIELD_FORMATS] = {}

        files = None
        format_key = None

        files = File.query.filter_by(aip_id=aip.id, file_type=FileType.original)
        if not original_files:
            files = File.query.filter_by(aip_id=aip.id, file_type=FileType.preservation)

        for file_ in files:
            try:
                format_key = file_.puid
            except AttributeError:
                format_key = file_.file_format
            if format_key is None:
                continue

            formats[format_key] = file_.file_format
            if file_.format_version:
                formats[format_key] = f"{file_.file_format} {file_.format_version}"

            size = aip_info[fields.FIELD_SIZE]
            try:
                aip_info[fields.FIELD_SIZE] = size + file_.size
            # TODO: Find out why size is sometimes None.
            except (AttributeError, TypeError):
                pass

            if format_key not in aip_info[fields.FIELD_FORMATS]:
                aip_info[fields.FIELD_FORMATS][format_key] = {}
                aip_info[fields.FIELD_FORMATS][format_key][fields.FIELD_COUNT] = 1
                try:
                    aip_info[fields.FIELD_FORMATS][format_key][fields.FIELD_VERSION] = (
                        file_.format_version
                    )
                    aip_info[fields.FIELD_FORMATS][format_key][fields.FIELD_NAME] = (
                        file_.file_format
                    )
                except AttributeError:
                    pass
            else:
                count = aip_info[fields.FIELD_FORMATS][format_key][fields.FIELD_COUNT]
                aip_info[fields.FIELD_FORMATS][format_key][fields.FIELD_COUNT] = (
                    count + 1
                )
        report[fields.FIELD_AIPS].append(aip_info)

    report[fields.FIELD_FORMATS] = formats

    return report


def derivative_overview(storage_service_id, storage_location_id=None):
    """Return a summary of derivatives across AIPs with a mapping
    created between the original format and the preservation copy.
    """
    report = report_dict(storage_service_id, storage_location_id)

    aips = AIP.query.filter_by(storage_service_id=storage_service_id)
    if storage_location_id:
        aips = aips.filter_by(storage_location_id=storage_location_id)
    aips = aips.all()
    all_aips = []
    for aip in aips:
        if not aip.preservation_file_count > 0:
            continue

        aip_report = {}
        aip_report[fields.FIELD_TRANSFER_NAME] = aip.transfer_name
        aip_report[fields.FIELD_UUID] = aip.uuid
        aip_report[fields.FIELD_FILE_COUNT] = aip.original_file_count
        aip_report[fields.FIELD_DERIVATIVE_COUNT] = aip.preservation_file_count
        aip_report[fields.FIELD_RELATED_PAIRING] = []

        original_files = File.query.filter_by(
            aip_id=aip.id, file_type=FileType.original
        )
        for original_file in original_files:
            preservation_derivative = File.query.filter_by(
                file_type=FileType.preservation, original_file_id=original_file.id
            ).first()

            if preservation_derivative is None:
                continue

            file_derivative_pair = {}
            file_derivative_pair[fields.FIELD_DERIVATIVE_UUID] = (
                preservation_derivative.uuid
            )
            file_derivative_pair[fields.FIELD_ORIGINAL_UUID] = original_file.uuid
            original_format_version = original_file.format_version
            if original_format_version is None:
                original_format_version = ""
            file_derivative_pair[fields.FIELD_ORIGINAL_FORMAT] = (
                f"{original_file.file_format} {original_format_version} ({original_file.puid})"
            )
            file_derivative_pair[fields.FIELD_DERIVATIVE_FORMAT] = (
                f"{preservation_derivative.file_format}"
            )
            aip_report[fields.FIELD_RELATED_PAIRING].append(file_derivative_pair)

        all_aips.append(aip_report)

    report[fields.FIELD_ALL_AIPS] = all_aips

    return report



import datetime
import uuid

import pytest
from flask import current_app

from AIPscan.models import File
from AIPscan.models import FileType
from AIPscan.Reporter.report_aips_by_puid import get_format_string_from_puid

EXPECTED_CSV_ORIGINAL = b"AIP Name,UUID,Count,Size,Size (bytes)\r\nTest AIP,111111111111-1111-1111-11111111,1,1.0 kB,1000\r\n"
EXPECTED_CSV_PRESERVATION = b"AIP Name,UUID,Count,Size,Size (bytes)\r\nTest AIP,111111111111-1111-1111-11111111,1,2.0 kB,2000\r\n"

FILE_WITH_FORMAT_ONLY = File(
    uuid=uuid.uuid4(),
    name="test.txt",
    size=12345,
    aip_id=2,
    file_type=FileType.original,
    file_format="Plain Text File",
    puid="x-fmt/111",
    filepath="/path/to/file.txt",
    date_created=datetime.datetime.now(),
    checksum_type="md5",
    checksum_value="anotherfakemd5",
)

FILE_WITH_FORMAT_AND_VERSION = File(
    uuid=uuid.uuid4(),
    name="test.pdf",
    size=12345678,
    aip_id=1,
    file_type=FileType.preservation,
    file_format="Acrobat PDF/A - Portable Document Format",
    format_version="1b",
    puid="fmt/354",
    filepath="/path/to/test.pdf",
    date_created=datetime.datetime.now(),
    checksum_type="md5",
    checksum_value="yetanotherfakemd5",
    original_file_id=1,
)

FILE_WITH_NO_FORMAT = File(
    uuid=uuid.uuid4(),
    name="test.txt",
    size=12345,
    aip_id=2,
    file_type=FileType.original,
    file_format=None,
    puid="x-fmt/111",
    filepath="/path/to/file.txt",
    date_created=datetime.datetime.now(),
    checksum_type="md5",
    checksum_value="anotherfakemd5",
)


@pytest.mark.parametrize(
    "puid, mock_file, expected_return_value",
    [
        ("x-fmt/111", FILE_WITH_FORMAT_ONLY, "Plain Text File"),
        (
            "fmt/354",
            FILE_WITH_FORMAT_AND_VERSION,
            "Acrobat PDF/A - Portable Document Format (1b)",
        ),
        ("x-fmt/111", FILE_WITH_NO_FORMAT, None),
        ("fmt/123", None, None),
    ],
)
def test_get_format_name_and_version_from_puid(
    app_instance, mocker, puid, mock_file, expected_return_value
):
    """Test that helper function returns expected string or None."""
    mock_get_file = mocker.patch("sqlalchemy.orm.query.Query.first")
    mock_get_file.return_value = mock_file
    assert expected_return_value == get_format_string_from_puid(puid)


@pytest.mark.parametrize(
    "original_files",
    [
        # Original files.
        ("True"),
        # Preservation files.
        ("False"),
    ],
)
def test_aips_by_puid(app_with_populated_files, original_files):
    """Test that report template renders."""
    with current_app.test_client() as test_client:
        response = test_client.get(
            f"/reporter/aips_by_puid/?amss_id=1&puid=fmt/353&original_files={original_files}"
        )
        assert response.status_code == 200


@pytest.mark.parametrize(
    "original_files,expected_csv",
    [
        # Original files.
        ("True", EXPECTED_CSV_ORIGINAL),
        # Preservation files.
        ("False", EXPECTED_CSV_PRESERVATION),
    ],
)
def test_aips_by_puid_csv(app_with_populated_files, original_files, expected_csv):
    """Test CSV export."""
    with current_app.test_client() as test_client:
        response = test_client.get(
            f"/reporter/aips_by_puid/?amss_id=1&puid=fmt/353&original_files={original_files}&csv=True"
        )
        assert response.status_code == 200
        assert (
            response.headers["Content-Disposition"]
            == "attachment; filename=aips_by_puid_fmt/353.csv"
        )
        assert response.mimetype == "text/csv"
        assert response.data == expected_csv



from flask import url_for


class NavBar:
    sections: list
    route_map: dict

    def __init__(self):
        self.sections = []
        self.route_map = {}

    def add_section(self, label, route):
        self.sections.append(NavBarSection(label, route, self))

    def map_route(self, route, label):
        self.route_map[route] = label


class NavBarSection:
    label: str
    route: str
    nav_bar: NavBar

    def __init__(self, label, route, nav_bar):
        self.label = label  # Navigation link text
        self.route = route  # Route ("transfer.index" for example)
        self.nav_bar = nav_bar  # Parent NavBar (Needed to access route map)

    def get_url(self):
        return url_for(self.route)

    def is_active(self, request):
        # Don't check if request isn't for a static asset
        request_rule = request.url_rule

        if request_rule is None:
            return False

        # Check for arbitrary mapping of request route to nav bar section
        if request_rule.endpoint in self.nav_bar.route_map:
            return self.nav_bar.route_map[request_rule.endpoint] == self.label

        request_blueprint = self.blueprint_of(request_rule.endpoint)

        if request_rule.endpoint == self.route:
            # A section's default route has been requested
            return True
        elif not self.blueprint_has_mapped_route(request_blueprint):
            # A route within a section representing an entire Blueprint may have been requested
            section_route_blueprint = self.blueprint_of(self.route)

            # Found Blueprint-level match between section and request
            return request_blueprint == section_route_blueprint

    def blueprint_has_mapped_route(self, blueprint):
        for mapped_route in self.nav_bar.route_map:
            route_map_blueprint = self.blueprint_of(mapped_route)

            if route_map_blueprint == blueprint:
                return True

        return False

    def blueprint_of(self, route):
        return route.split(".")[0]



from functools import wraps

from flask import abort
from flask import render_template
from flask import request

from AIPscan import db


def confirm_required(model, id_argument, prompt, action, cancel_route):
    def decorator(f):
        @wraps(f)
        def decorated_function(*args, **kwargs):
            instance = db.session.get(model, kwargs[id_argument])

            if not instance:
                abort(404)

            if request.args.get("confirm"):
                return f(*args, **kwargs)

            return render_template(
                "confirm.html",
                prompt=prompt,
                action=action,
                cancel_route=cancel_route,
            )

        return decorated_function

    return decorator



from flask import Blueprint
from flask import redirect
from flask import url_for

home = Blueprint("home", __name__)


@home.route("/", methods=["GET"])
def index():
    """Define handling for application's / route."""
    return redirect(url_for("aggregator.ss_default"))



import hashlib

import pytest

from AIPscan.helpers import filesizeformat
from AIPscan.helpers import stream_write_and_hash


@pytest.mark.parametrize(
    "byte_size,binary,expected_result",
    [
        # Test non-binary results.
        (100, False, "100 Bytes"),
        (1000, False, "1.0 kB"),
        (1000000, False, "1.0 MB"),
        (1000000000, False, "1.0 GB"),
        (1000000000000, False, "1.0 TB"),
        # Test binary results.
        (100, True, "100 Bytes"),
        (1000, True, "1000 Bytes"),
        (1000000, True, "976.6 KiB"),
        (1000000000, True, "953.7 MiB"),
        (1000000000000, True, "931.3 GiB"),
        # Test real-life values.
        (396245, False, "396.2 kB"),
        (396245, True, "387.0 KiB"),
        (125121042160, False, "125.1 GB"),
        (125121042160, True, "116.5 GiB"),
    ],
)
def test_filesizeformat(byte_size, binary, expected_result):
    assert filesizeformat(byte_size, binary=binary) == expected_result


class _FakeStreamingResponse:
    def __init__(self, chunks):
        self._chunks = chunks

    def iter_content(self, chunk_size=None):
        yield from self._chunks


def test_stream_write_and_hash(tmp_path):
    payload = b"test payload for mets file"
    chunks = [payload[:5], payload[5:-1], payload[-1:], b""]
    response = _FakeStreamingResponse(chunks)

    destination = tmp_path / "mets.xml"

    digest = stream_write_and_hash(response, destination)

    assert destination.read_bytes() == payload
    assert digest == hashlib.sha256(payload).hexdigest()


def test_stream_write_and_hash_with_iterable(tmp_path):
    payload = b"chunked bytes"
    chunks = [payload[:6], payload[6:]]

    destination = tmp_path / "payload.bin"

    digest = stream_write_and_hash(chunks, destination)

    assert destination.read_bytes() == payload
    assert digest == hashlib.sha256(payload).hexdigest()



"""Code shared across reporting modules but not outside of reporting."""

import csv
from datetime import timedelta
from io import StringIO

from flask import make_response
from natsort import natsorted

from AIPscan.Data import fields
from AIPscan.helpers import filesizeformat


def sort_puids(puids):
    """Return PUIDs sorted in natural sorting order.

    Python sorts strings containining integers in ASCII order - i.e.
    ["fmt/1", "fmt/10", "fmt/2"]. Since this is different than what end
    users expect, this helper uses the natsort library to reorder PUIDs
    into natural sort order - i.e. ["fmt/1", "fmt/2", "fmt/10"].

    :param puids: List of PUIDs

    :returns: Sorted list of PUIDs
    """
    return natsorted(puids)


def translate_headers(headers, add_bytes_column=False):
    """Translate headers from something machine readable to something
    more user friendly and translatable.
    """
    field_lookup = {
        fields.FIELD_AIP_NAME: "AIP Name",
        fields.FIELD_AIPS: "AIPs",
        fields.FIELD_AIP_SIZE: "AIP Size",
        fields.FIELD_AIP_UUID: "AIP UUID",
        fields.FIELD_ALL_AIPS: "All AIPs",
        fields.FIELD_COUNT: "Count",
        fields.FIELD_CREATED_DATE: "Created Date",
        fields.FIELD_DATE_START: "Start Date",
        fields.FIELD_DATE_END: "End Date",
        fields.FIELD_DURATION: "Duration",
        fields.FIELD_DERIVATIVE_COUNT: "Derivative Count",
        fields.FIELD_DERIVATIVE_FILE: "Derivative File",
        fields.FIELD_DERIVATIVE_FORMAT: "Derivative Format",
        fields.FIELD_DERIVATIVE_UUID: "Derivative UUID",
        fields.FIELD_FILE_COUNT: "File Count",
        fields.FIELD_FILE_TYPE: "Type",
        fields.FIELD_FILENAME: "Filename",
        fields.FIELD_FORMAT: "Format",
        fields.FIELD_FORMATS: "Formats",
        fields.FIELD_ID: "ID",
        fields.FIELD_NAME: "Name",
        fields.FIELD_ORIGINAL_FILE: "Original File",
        fields.FIELD_ORIGINAL_FORMAT: "Original Format",
        fields.FIELD_ORIGINAL_NAME: "Original Name",
        fields.FIELD_ORIGINAL_PUID: "Original PUID",
        fields.FIELD_ORIGINAL_UUID: "Original UUID",
        fields.FIELD_ORIGINAL_VERSION: "Original Version",
        fields.FIELD_PUID: "PUID",
        fields.FIELD_RELATED_PAIRING: "Related Pairing",
        fields.FIELD_SIZE: "Size",
        fields.FIELD_STORAGE_LOCATION: "Location",
        fields.FIELD_STORAGE_NAME: "Storage Service Name",
        fields.FIELD_TRANSFER_NAME: "Transfer Name",
        fields.FIELD_USER: "User",
        fields.FIELD_VERSION: "Version",
    }

    # Attempt to add an additional header representing a column containing size
    # expressed as a number of bytes, rather than in human-readable form, so
    # rows can more easily be sorted by size
    if add_bytes_column:
        headers = (
            headers.copy()
        )  # So we don't change the list object passed to this function

        # Handle the two standard size columns
        for header in [fields.FIELD_AIP_SIZE, fields.FIELD_SIZE]:
            # If size header is found then insert another for the size in bytes afer it
            if header in headers:
                bytes_header = field_lookup[header] + " (bytes)"
                headers.insert(headers.index(header) + 1, bytes_header)

    return [field_lookup.get(header, header) for header in headers]


def _remove_primary_keys(dictionary):
    """Remove AIPscan primary keys from dictionary."""
    PK_FIELDS = (fields.FIELD_ID, fields.FIELD_AIP_ID)
    for field in PK_FIELDS:
        try:
            dictionary.pop(field, None)
        except AttributeError:
            pass


def format_size_for_csv(rows):
    """Return data prepared for CSV file.

    :param rows: Data assembled for tabular report (list of dicts)

    :returns: rows with formatted size field (list of dicts)
    """
    edited_rows = []

    for row in rows:
        # Add size in bytes after original size column
        row_key_list = list(row.keys())

        if fields.FIELD_SIZE in row_key_list:
            size_position = row_key_list.index(fields.FIELD_SIZE) + 1
            row_items = list(row.items())

            row_items.insert(
                size_position, (fields.FIELD_SIZE_BYTES, row[fields.FIELD_SIZE])
            )
            row = dict(row_items)

            # Format original size column
            row[fields.FIELD_SIZE] = filesizeformat(row[fields.FIELD_SIZE])

        edited_rows.append(row)

    return edited_rows


def download_csv(headers, rows, filename="report.csv"):
    """Write CSV from string and send it as an attachment.

    :param headers: Row headers (list of str)
    :param rows: Data to write to CSV, returned from Data endpoint (list of dicts)
    :param filename: CSV filename (str)
    """
    string_io = StringIO()
    writer = csv.writer(string_io)
    writer.writerow(headers)
    for row in rows:
        _remove_primary_keys(row)
        writer.writerow(row.values())
    response = make_response(string_io.getvalue())
    response.headers["Content-Disposition"] = f"attachment; filename={filename}"
    response.mimetype = "text/csv"
    return response


def get_display_end_date(end_date):
    """Format end date to display.

    A day is added to end_date by parse_datetime_bound to facilitate date
    comparison in SQL queries. Here we remove that extra day before displaying
    the date in a report.

    :param end_date: End date (datetime.datetime object)

    :return: End date minus one day (datetime.datetime object)
    """
    return end_date - timedelta(days=1)


def calculate_paging_window(pagination):
    first_item = ((pagination.page - 1) * pagination.per_page) + 1
    last_item = pagination.page * pagination.per_page

    return first_item, min(last_item, pagination.total)


def remove_dict_none_values(values):
    return {index: "" if value is None else value for (index, value) in values.items()}



"""Request parameters."""

AIP_UUID = "aip_uuid"
CSV = "csv"
CUMULATIVE = "cumulative"
FILE_FORMAT = "file_format"
FILE_TYPE = "file_type"
LIMIT = "limit"
METRIC = "metric"
ORIGINAL_FILES = "original_files"
PUID = "puid"
STORAGE_LOCATION_ID = "storage_location"
STORAGE_SERVICE_ID = "amss_id"
PAGE = "page"

START_DATE = "start_date"
END_DATE = "end_date"



import os
from importlib.metadata import PackageNotFoundError
from importlib.metadata import version as _metadata_version

try:
    __version__ = _metadata_version("AIPscan")
except PackageNotFoundError:
    __version__ = "0.0.0+unknown"

__all__ = ["__version__"]

from flask import Flask
from flask import render_template
from flask_migrate import Migrate
from flask_sqlalchemy import SQLAlchemy

from AIPscan.celery import configure_celery
from AIPscan.config import CONFIGS
from AIPscan.navbar import NavBar

db = SQLAlchemy()
migrate = Migrate()


def create_app(config_name=None):
    """Flask app factory, returns app instance.

    If no config_name is passed (i.e. config_name is None), honor the
    ``FLASK_CONFIG`` environment variable; otherwise default to "default".
    """
    if config_name is None:
        config_name = os.environ.get("FLASK_CONFIG", "default")
    config = CONFIGS.get(config_name)

    app = Flask(__name__)
    app.config.from_object(config)

    # Set up the secret key.
    secret = os.getenv("SECRET_KEY") or app.config.get("SECRET_KEY")
    if app.debug or app.testing:
        app.config["SECRET_KEY"] = secret or "dev"
    else:
        if not secret or not str(secret).strip():
            raise RuntimeError(
                "SECRET_KEY must be set to a non-empty value when running in production."
            )
        app.config["SECRET_KEY"] = secret

    migrations_dir = os.path.abspath(
        os.path.join(os.path.dirname(__file__), "migrations")
    )
    migrate.init_app(app, db, directory=migrations_dir, compare_type=True)

    app.logger.info("Starting AIPscan application... (config=%s)", config_name)

    with app.app_context():
        from AIPscan.Aggregator.views import aggregator
        from AIPscan.API.views import api
        from AIPscan.Home.views import home
        from AIPscan.Reporter.views import reporter
        from AIPscan.User.views import user

        app.register_blueprint(aggregator, url_prefix="/aggregator")
        app.register_blueprint(reporter, url_prefix="/reporter")
        app.register_blueprint(user, url_prefix="/user")
        app.register_blueprint(api)
        app.register_blueprint(home)

        # Ensure all SQLAlchemy models are registered before schema creation.
        from AIPscan import models  # noqa: F401

        db.init_app(app)
        configure_celery(app)

        from AIPscan.cli import register_cli

        register_cli(app)

        # Define navigation bar sections and route-to-section mapping (needed
        # given that the "AIPs" and "Reports" sections are in the same Blueprint)
        navbar = NavBar()
        navbar.add_section(
            "Archivematica Storage Services", "aggregator.storage_services"
        )
        navbar.add_section("AIPs", "reporter.view_aips")
        navbar.add_section("Reports", "reporter.reports")

        navbar.map_route("reporter.view_aip", "AIPs")
        navbar.map_route("reporter.view_file", "AIPs")

        # Inject navigation bar into templates
        @app.context_processor
        def inject_navbar():
            return dict(navbar=navbar)

        # Set up 404 handling
        @app.errorhandler(404)
        def page_not_found(e):
            return render_template("error/404.html"), 404

        return app



import os
from importlib import resources

DEFAULT_AIPSCAN_DB = (
    "mysql+pymysql://aipscan:demo@127.0.0.1:3406/aipscan?charset=utf8mb4"
)
DEFAULT_TEST_DB = (
    "mysql+pymysql://aipscan:demo@127.0.0.1:3406/aipscan_test?charset=utf8mb4"
)

DEFAULT_TYPESENSE_HOST = "localhost"
DEFAULT_TYPESENSE_PORT = "8108"
DEFAULT_TYPESENSE_API_KEY = None
DEFAULT_TYPESENSE_PROTOCOL = "http"
DEFAULT_TYPESENSE_TIMEOUT_SECONDS = "30"
DEFAULT_TYPESENSE_COLLECTION_PREFIX = "aipscan_"
DEFAULT_AGGREGATOR_DOWNLOAD_ROOT = os.fspath(
    resources.files(__package__).joinpath("Aggregator", "downloads")
)


class Config:
    SECRET_KEY = None
    DEBUG = False
    TESTING = False

    SQLALCHEMY_DATABASE_URI = os.getenv("SQLALCHEMY_DATABASE_URI", DEFAULT_AIPSCAN_DB)
    SQLALCHEMY_TRACK_MODIFICATIONS = False
    SQLALCHEMY_ECHO = False
    SQLALCHEMY_ENGINE_OPTIONS = {
        "pool_pre_ping": True,
    }

    TYPESENSE_HOST = os.getenv("TYPESENSE_HOST", DEFAULT_TYPESENSE_HOST)
    TYPESENSE_PORT = os.getenv("TYPESENSE_PORT", DEFAULT_TYPESENSE_PORT)
    TYPESENSE_API_KEY = os.getenv("TYPESENSE_API_KEY", DEFAULT_TYPESENSE_API_KEY)
    TYPESENSE_PROTOCOL = os.getenv("TYPESENSE_PROTOCOL", DEFAULT_TYPESENSE_PROTOCOL)
    TYPESENSE_TIMEOUT_SECONDS = os.getenv(
        "TYPESENSE_TIMEOUT_SECONDS", DEFAULT_TYPESENSE_TIMEOUT_SECONDS
    )
    TYPESENSE_COLLECTION_PREFIX = os.getenv(
        "TYPESENSE_COLLECTION_PREFIX", DEFAULT_TYPESENSE_COLLECTION_PREFIX
    )
    AGGREGATOR_DOWNLOAD_ROOT = os.getenv(
        "AGGREGATOR_DOWNLOAD_ROOT", DEFAULT_AGGREGATOR_DOWNLOAD_ROOT
    )


class DevelopmentConfig(Config):
    DEBUG = True


class TestConfig(Config):
    TESTING = True
    SQLALCHEMY_DATABASE_URI = os.getenv("SQLALCHEMY_TEST_DATABASE_URI", DEFAULT_TEST_DB)


CONFIGS = {"dev": DevelopmentConfig, "test": TestConfig, "default": Config}



"""Data endpoints optimized for reports in the Reporter blueprint."""

from datetime import datetime
from operator import itemgetter

from dateutil.rrule import DAILY
from dateutil.rrule import rrule
from sqlalchemy import case

from AIPscan import db
from AIPscan.Data import fields
from AIPscan.Data import get_storage_service_name
from AIPscan.Data import report_dict
from AIPscan.models import AIP
from AIPscan.models import Event
from AIPscan.models import File
from AIPscan.models import FileType
from AIPscan.models import StorageLocation
from AIPscan.models import StorageService

VALID_FILE_TYPES = set(item.value for item in FileType)


def _get_username(agent_string):
    """Retrieve username from the standard agent string stored in the
    database, normally formatted as:

        * username="test", first_name="", last_name=""
    """
    USERNAME = "username="
    return agent_string.split(",", 1)[0].replace(USERNAME, "").replace('"', "")


def _formats_count_query(
    storage_service_id, start_date, end_date, storage_location_id=None
):
    """Fetch information from database on file formats.

    :param storage_service_id: Storage Service ID (int)
    :param start_date: Inclusive AIP creation start date
        (datetime.datetime object)
    :param end_date: Inclusive AIP creation end date
        (datetime.datetime object)
    :param storage_location_id: Storage Location ID (int)

    :returns: SQLAlchemy query results
    """
    FILE_FORMAT = "file_format"
    FILE_COUNT = "file_count"
    FILE_SIZE = "total_size"

    results = (
        db.session.query(
            File.file_format.label(FILE_FORMAT),
            db.func.count(File.id).label(FILE_COUNT),
            db.func.sum(File.size).label(FILE_SIZE),
        )
        .join(AIP)
        .join(StorageLocation)
        .join(StorageService)
        .filter(StorageService.id == storage_service_id)
        .filter(File.file_type == FileType.original.value)
        .filter(AIP.create_date >= start_date)
        .filter(AIP.create_date < end_date)
        .group_by(File.file_format)
        .order_by(db.func.count(File.id).desc(), db.func.sum(File.size).desc())
    )
    if storage_location_id:
        results = results.filter(StorageLocation.id == storage_location_id)
    return results


def formats_count(storage_service_id, start_date, end_date, storage_location_id=None):
    """Return a summary of file formats in Storage Service.

    :param storage_service_id: Storage Service ID (int)
    :param start_date: Inclusive AIP creation start date
        (datetime.datetime object)
    :param end_date: Inclusive AIP creation end date
        (datetime.datetime object)
    :param storage_location_id: Storage Location ID (int)

    :returns: "report" dict containing following fields:
        report["StorageName"]: Name of Storage Service queried
        report["Formats"]: List of results ordered desc by count and size
    """
    report = report_dict(storage_service_id, storage_location_id)
    report[fields.FIELD_FORMATS] = []

    formats = _formats_count_query(
        storage_service_id, start_date, end_date, storage_location_id
    )

    for format_ in formats:
        format_info = {}

        format_info[fields.FIELD_FORMAT] = format_.file_format
        format_info[fields.FIELD_COUNT] = format_.file_count
        format_info[fields.FIELD_SIZE] = 0
        if format_.total_size is not None:
            format_info[fields.FIELD_SIZE] = format_.total_size

        report[fields.FIELD_FORMATS].append(format_info)

    return report


def _format_versions_count_query(
    storage_service_id, start_date, end_date, storage_location_id
):
    """Fetch information from database on format versions.

    :param storage_service_id: Storage Service ID (int)
    :param start_date: Inclusive AIP creation start date
        (datetime.datetime object)
    :param end_date: Inclusive AIP creation end date
        (datetime.datetime object)
    :param storage_location_id: Storage Location ID (int)

    :returns: SQLAlchemy query results
    """
    PUID = "puid"
    FILE_FORMAT = "file_format"
    FORMAT_VERSION = "format_version"
    FILE_COUNT = "file_count"
    FILE_SIZE = "total_size"

    results = (
        db.session.query(
            File.puid.label(PUID),
            File.file_format.label(FILE_FORMAT),
            File.format_version.label(FORMAT_VERSION),
            db.func.count(File.id).label(FILE_COUNT),
            db.func.sum(File.size).label(FILE_SIZE),
        )
        .join(AIP)
        .join(StorageLocation)
        .join(StorageService)
        .filter(StorageService.id == storage_service_id)
        .filter(File.file_type == FileType.original.value)
        .filter(AIP.create_date >= start_date)
        .filter(AIP.create_date < end_date)
        .group_by(File.puid, File.file_format, File.format_version)
        .order_by(db.func.count(File.id).desc(), db.func.sum(File.size).desc())
    )
    if storage_location_id:
        results = results.filter(StorageLocation.id == storage_location_id)
    return results


def format_versions_count(
    storage_service_id, start_date, end_date, storage_location_id=None
):
    """Return a summary of format versions in Storage Service.

    :param storage_service_id: Storage Service ID (int)
    :param start_date: Inclusive AIP creation start date
        (datetime.datetime object)
    :param end_date: Inclusive AIP creation end date
        (datetime.datetime object)
    :param storage_location_id: Storage Location ID (int)

    :returns: "report" dict containing following fields:
        report["StorageName"]: Name of Storage Service queried
        report["FormatVersions"]: List of result files ordered desc by size
    """
    report = report_dict(storage_service_id, storage_location_id)
    report[fields.FIELD_FORMAT_VERSIONS] = []

    versions = _format_versions_count_query(
        storage_service_id, start_date, end_date, storage_location_id
    )

    for version in versions:
        version_info = {}

        version_info[fields.FIELD_PUID] = version.puid
        version_info[fields.FIELD_FORMAT] = version.file_format
        try:
            version_info[fields.FIELD_VERSION] = version.format_version
        except AttributeError:
            pass
        version_info[fields.FIELD_COUNT] = version.file_count
        version_info[fields.FIELD_SIZE] = 0
        if version.total_size is not None:
            version_info[fields.FIELD_SIZE] = version.total_size

        report[fields.FIELD_FORMAT_VERSIONS].append(version_info)

    return report


def _largest_files_query(
    storage_service_id, start_date, end_date, storage_location_id, file_type, limit
):
    """Fetch file information from database for largest files query

    This is separated into its own helper function to aid in testing.
    """
    files = (
        File.query.join(AIP)
        .join(StorageLocation)
        .join(StorageService)
        .filter(StorageService.id == storage_service_id)
        .filter(AIP.create_date >= start_date)
        .filter(AIP.create_date < end_date)
        .order_by(File.size.desc())
    )
    if file_type is not None and file_type in VALID_FILE_TYPES:
        files = files.filter(File.file_type == file_type)
    if storage_location_id:
        files = files.filter(StorageLocation.id == storage_location_id)
    return files.limit(limit)


def largest_files(
    storage_service_id,
    start_date,
    end_date,
    storage_location_id=None,
    file_type=None,
    limit=20,
):
    """Return a summary of the largest files in a given Storage Service

    :param storage_service_id: Storage Service ID
    :param start_date: Inclusive AIP creation start date
        (datetime.datetime object)
    :param end_date: Inclusive AIP creation end date
        (datetime.datetime object)
    :param storage_location_id: Storage Location ID (int)
    :param file_type: Optional filter for type of file to return
        (acceptable values are "original" or "preservation")
    :param limit: Upper limit of number of results to return

    :returns: "report" dict containing following fields:
        report["StorageName"]: Name of Storage Service queried
        report["Files"]: List of result files ordered desc by size
    """
    report = report_dict(storage_service_id, storage_location_id)
    report[fields.FIELD_FILES] = []

    files = _largest_files_query(
        storage_service_id, start_date, end_date, storage_location_id, file_type, limit
    )

    for file_ in files:
        file_info = {}

        file_info[fields.FIELD_ID] = file_.id
        file_info[fields.FIELD_UUID] = file_.uuid
        file_info[fields.FIELD_NAME] = file_.name
        try:
            file_info[fields.FIELD_SIZE] = int(file_.size)
        except TypeError:
            file_info[fields.FIELD_SIZE] = 0
        file_info[fields.FIELD_AIP_ID] = file_.aip_id
        file_info[fields.FIELD_FILE_TYPE] = file_.file_type.value

        try:
            file_info[fields.FIELD_FORMAT] = file_.file_format
        except AttributeError:
            pass
        try:
            file_info[fields.FIELD_VERSION] = file_.format_version
        except AttributeError:
            pass
        try:
            file_info[fields.FIELD_PUID] = file_.puid
        except AttributeError:
            pass

        matching_aip = db.session.get(AIP, file_.aip_id)
        if matching_aip is not None:
            file_info[fields.FIELD_AIP_NAME] = matching_aip.transfer_name
            file_info[fields.FIELD_AIP_UUID] = matching_aip.uuid

        report[fields.FIELD_FILES].append(file_info)

    return report


def _largest_aips_query(
    storage_service_id, start_date, end_date, storage_location_id, limit
):
    """Fetch information from database for largest AIPs query."""
    aips = (
        AIP.query.join(StorageLocation)
        .join(StorageService)
        .filter(StorageService.id == storage_service_id)
        .filter(AIP.create_date >= start_date)
        .filter(AIP.create_date < end_date)
    )
    if storage_location_id:
        aips = aips.filter(StorageLocation.id == storage_location_id)
    return aips.order_by(AIP.size.desc()).limit(limit)


def largest_aips(
    storage_service_id, start_date, end_date, storage_location_id=None, limit=20
):
    """Return a summary of the largest AIPs in a given Storage Service

    :param storage_service_id: Storage Service ID
    :param start_date: Inclusive AIP creation start date
        (datetime.datetime object)
    :param end_date: Inclusive AIP creation end date
        (datetime.datetime object)
    :param storage_location_id: Storage Location ID (int)
    :param limit: Upper limit of number of results to return

    :returns: "report" dict containing following fields:
        report["StorageName"]: Name of Storage Service queried
        report["AIPs"]: List of result AIPs ordered desc by size
    """
    report = report_dict(storage_service_id, storage_location_id)
    report[fields.FIELD_AIPS] = []

    aips = _largest_aips_query(
        storage_service_id, start_date, end_date, storage_location_id, limit
    )

    for aip in aips:
        aip_info = {}

        aip_info[fields.FIELD_NAME] = aip.transfer_name
        aip_info[fields.FIELD_UUID] = aip.uuid
        aip_info[fields.FIELD_SIZE] = aip.size
        aip_info[fields.FIELD_FILE_COUNT] = aip.original_file_count

        report[fields.FIELD_AIPS].append(aip_info)

    return report


def _query_aips_by_file_format_or_puid(
    storage_service_id,
    storage_location_id,
    search_string,
    original_files=True,
    file_format=True,
):
    """Fetch information on all AIPs with given format or PUID from db.

    :param storage_service_id: Storage Service ID (int)
    :param storage_location_id: Storage Location ID (int)
    :param search_string: File format or PUID (str)
    :param original_files: Flag indicating whether returned data
        describes original (default) or preservation files (bool)
    :param file_format: Flag indicating whether to filter on file
        format (default) or PUID (bool)

    :returns: SQLAlchemy query results
    """
    AIP_ID = "id"
    TRANSFER_NAME = "name"
    AIP_UUID = "uuid"
    FILE_COUNT = "file_count"
    FILE_SIZE = "total_size"
    aips = (
        db.session.query(
            AIP.id.label(AIP_ID),
            AIP.transfer_name.label(TRANSFER_NAME),
            AIP.uuid.label(AIP_UUID),
            db.func.count(File.id).label(FILE_COUNT),
            db.func.sum(File.size).label(FILE_SIZE),
        )
        .join(File)
        .join(StorageService)
        .filter(StorageService.id == storage_service_id)
        .group_by(AIP.id)
        .order_by(db.func.count(File.id).desc(), db.func.sum(File.size).desc())
    )
    if storage_location_id:
        aips = aips.filter(AIP.storage_location_id == storage_location_id)

    if original_files is False:
        aips = aips.filter(File.file_type == FileType.preservation.value)
    else:
        aips = aips.filter(File.file_type == FileType.original.value)

    if file_format:
        return aips.filter(File.file_format == search_string)
    return aips.filter(File.puid == search_string)


def _aips_by_file_format_or_puid(
    storage_service_id,
    storage_location_id,
    search_string,
    original_files=True,
    file_format=True,
):
    """Return overview of all AIPs containing original files in format

    :param storage_service_id: Storage Service ID (int)
    :param storage_location_id: Storage Location ID (int)
    :param search_string: File format name or PUID (str)
    :param original_files: Flag indicating whether returned data
        data describes original (default) or preservation files (bool)
    :param file_format: Flag indicating whether to filter on file
        file format (default) or PUID (bool)

    :returns: "report" dict containing following fields:
        report["StorageName"]: Name of Storage Service queried
        report["AIPs"]: List of result AIPs ordered desc by count
    """
    report = report_dict(storage_service_id, storage_location_id)

    if file_format:
        report[fields.FIELD_FORMAT] = search_string
    else:
        report[fields.FIELD_PUID] = search_string

    report[fields.FIELD_AIPS] = []
    results = _query_aips_by_file_format_or_puid(
        storage_service_id,
        storage_location_id,
        search_string,
        original_files,
        file_format,
    )
    for result in results:
        aip_info = {}

        aip_info[fields.FIELD_ID] = result.id
        aip_info[fields.FIELD_AIP_NAME] = result.name
        aip_info[fields.FIELD_UUID] = result.uuid
        aip_info[fields.FIELD_COUNT] = result.file_count
        aip_info[fields.FIELD_SIZE] = result.total_size

        report[fields.FIELD_AIPS].append(aip_info)

    return report


def aips_by_file_format(
    storage_service_id, file_format, original_files=True, storage_location_id=None
):
    """Return overview of AIPs containing original files in format.

    :param storage_service_id: Storage Service ID (int)
    :param file_format: File format name (str)
    :param original_files: Flag indicating whether returned data
        describes original (default) or preservation files (bool)
    :param storage_location_id: Storage Location ID (int)

    :returns: Report dict provided by _aips_by_file_format_or_puid
    """
    return _aips_by_file_format_or_puid(
        storage_service_id=storage_service_id,
        storage_location_id=storage_location_id,
        search_string=file_format,
        original_files=original_files,
    )


def aips_by_puid(
    storage_service_id, puid, original_files=True, storage_location_id=None
):
    """Return overview of AIPs containing original files in format.

    :param storage_service_id: Storage Service ID (int)
    :param puid: PUID (str)
    :param original_files: Flag indicating whether returned data
        describes original (default) or preservation files (bool)
    :param storage_location_id: Storage Location ID (int)

    :returns: Report dict provided by _aips_by_file_format_or_puid
    """
    return _aips_by_file_format_or_puid(
        storage_service_id=storage_service_id,
        storage_location_id=storage_location_id,
        search_string=puid,
        original_files=original_files,
        file_format=False,
    )


def agents_transfers(
    storage_service_id, start_date, end_date, storage_location_id=None
):
    """Return overview of transfers and their agents.

    Includes simple statistics such as ingest start and finish times.

    :param storage_service_id: Storage Service ID (int)
    :param start_date: Inclusive AIP creation start date
        (datetime.datetime object)
    :param end_date: Inclusive AIP creation end date
        (datetime.datetime object)
    :param storage_location_id: Storage Location ID (int)

    :returns: "report" dict containing following fields:
        report["StorageName"]: Name of Storage Service queried
        report["StorageLocation"]: Name of storage location, if applicable
        report["Ingest"]: List of ingest event dictionaries
    """
    report = report_dict(storage_service_id, storage_location_id)

    ingests = []

    if not report[fields.FIELD_STORAGE_NAME]:
        # No storage service has been returned and so we have nothing
        # to return.
        report[fields.FIELD_INGESTS] = ingests
        return report

    aips = (
        AIP.query.filter_by(storage_service_id=storage_service_id)
        .filter(AIP.create_date >= start_date)
        .filter(AIP.create_date < end_date)
    )
    if storage_location_id:
        aips = aips.filter(AIP.storage_location_id == storage_location_id)
    aips = aips.all()

    EVENT_TYPE = "ingestion"
    AGENT_TYPE = "Archivematica user"

    for aip in aips:
        event = (
            db.session.query(Event)
            .join(File)
            .filter(File.aip_id == aip.id, Event.type == EVENT_TYPE)
            .first()
        )
        # This defensive check is necessary for now because of packages that
        # are deleted after extraction. See issue #104 for details.
        if event is None:
            continue
        log_line = {}
        log_line[fields.FIELD_AIP_UUID] = aip.uuid
        log_line[fields.FIELD_AIP_NAME] = aip.transfer_name
        log_line[fields.FIELD_INGEST_START_DATE] = str(event.date)
        log_line[fields.FIELD_INGEST_FINISH_DATE] = str(aip.create_date)
        for agent in event.event_agents:
            if agent.agent_type == AGENT_TYPE:
                log_line[fields.FIELD_USER] = _get_username(agent.agent_value)
        ingests.append(log_line)
    report[fields.FIELD_INGESTS] = ingests
    return report


def _preservation_derivatives_query(storage_service_id, storage_location_id, aip_uuid):
    """Fetch information on preservation derivatives from db.

    :param storage_service_id: Storage Service ID (int)
    :param storage_location_id: Storage Location ID (int)
    :param aip_uuid: AIP UUID (str)

    :returns: SQLAlchemy query results
    """
    files = (
        File.query.join(AIP)
        .join(StorageLocation)
        .join(StorageService)
        .filter(StorageService.id == storage_service_id)
        .filter(File.file_type == FileType.preservation)
        .order_by(AIP.uuid, File.file_format)
    )
    if storage_location_id:
        files = files.filter(StorageLocation.id == storage_location_id)
    if aip_uuid:
        files = files.filter(AIP.uuid == aip_uuid)
    return files


def preservation_derivatives(
    storage_service_id, storage_location_id=None, aip_uuid=None
):
    """Return details of preservation derivatives in Storage Service.

    This includes information about each preservation derivative, as well as
    its corresponding original file and AIP.

    :param storage_service_id: Storage Service ID (int)
    :param storage_location_id: Storage Location ID (int)
    :param aip_uuid: AIP UUID (str)

    :returns: "report" dict containing following fields:
        report["StorageName"]: Name of Storage Service queried
        report["Files"]: List of result files ordered desc by size
    """
    report = report_dict(storage_service_id, storage_location_id)
    report[fields.FIELD_FILES] = []

    files = _preservation_derivatives_query(
        storage_service_id, storage_location_id, aip_uuid
    )

    for file_ in files:
        file_info = {}

        file_info[fields.FIELD_AIP_UUID] = file_.aip.uuid
        file_info[fields.FIELD_AIP_NAME] = file_.aip.transfer_name

        file_info[fields.FIELD_ID] = file_.id
        file_info[fields.FIELD_UUID] = file_.uuid
        file_info[fields.FIELD_NAME] = file_.name
        file_info[fields.FIELD_FORMAT] = file_.file_format

        original_file = file_.original_file
        if original_file:
            file_info[fields.FIELD_ORIGINAL_UUID] = original_file.uuid
            file_info[fields.FIELD_ORIGINAL_NAME] = original_file.name
            file_info[fields.FIELD_ORIGINAL_FORMAT] = original_file.file_format
            file_info[fields.FIELD_ORIGINAL_VERSION] = ""
            try:
                file_info[fields.FIELD_ORIGINAL_VERSION] = original_file.format_version
            except AttributeError:
                pass
            file_info[fields.FIELD_ORIGINAL_PUID] = ""
            try:
                file_info[fields.FIELD_ORIGINAL_PUID] = original_file.puid
            except AttributeError:
                pass

        report[fields.FIELD_FILES].append(file_info)

    return report


def _get_storage_locations(storage_service_id):
    """Return queryset of locations in this Storage Service."""
    return StorageLocation.query.filter_by(storage_service_id=storage_service_id).all()


def _sort_storage_locations(unsorted_locations):
    """Sort list of location dictionaries by AIP count descending."""
    return sorted(unsorted_locations, key=itemgetter(fields.FIELD_AIPS), reverse=True)


def storage_locations(storage_service_id, start_date, end_date):
    """Return details of AIP store locations in Storage Service.

    :param storage_service_id: Storage Service ID (int)
    :param start_date: Inclusive AIP creation start date
        (datetime.datetime object)
    :param end_date: Inclusive AIP creation end date
        (datetime.datetime object)

    :returns: "report" dict containing following fields:
        report["StorageName"]: Name of Storage Service queried
        report["Locations"]: List of result locations ordered desc by size
    """
    report = {}
    report[fields.FIELD_STORAGE_NAME] = get_storage_service_name(storage_service_id)

    locations = _get_storage_locations(storage_service_id)

    unsorted_results = []

    for location in locations:
        loc_info = {}

        loc_info[fields.FIELD_ID] = location.id
        loc_info[fields.FIELD_UUID] = location.uuid
        loc_info[fields.FIELD_STORAGE_LOCATION] = location.description
        loc_info[fields.FIELD_AIPS] = location.aip_count(start_date, end_date)
        loc_info[fields.FIELD_SIZE] = location.aip_total_size(start_date, end_date)
        loc_info[fields.FIELD_FILE_COUNT] = location.file_count(start_date, end_date)

        unsorted_results.append(loc_info)

    report[fields.FIELD_LOCATIONS] = _sort_storage_locations(unsorted_results)

    return report


def _get_days_covered_by_date_range(storage_service_id, start_date, end_date):
    """Return days covered by date range as list of ISO values.

    First, replace datetime.min and datetime.max, if used, with more sensible
    values so that we do not end up returning every day since the beginning
    of time.

    :param storage_service_id: Storage Service ID (int)
    :param start_date: Inclusive AIP creation start date
        (datetime.datetime object)
    :param end_date: Inclusive AIP creation end date
        (datetime.datetime object)

    :returns: List of strings like "YYYY-MM-DD"
    """
    if start_date == datetime.min:
        storage_service = db.session.get(StorageService, storage_service_id)
        start_date = storage_service.earliest_aip_created
    if end_date >= datetime.now():
        end_date = datetime.now()

    return [
        dt.isoformat()[:10] for dt in rrule(DAILY, dtstart=start_date, until=end_date)
    ]


def storage_locations_usage_over_time(
    storage_service_id, start_date, end_date, cumulative=False
):
    """Return details of AIP store locations in Storage Service over time.

    :param storage_service_id: Storage Service ID (int)
    :param start_date: Inclusive AIP creation start date
        (datetime.datetime object)
    :param end_date: Exclusive upper bound for AIP creation timestamps
        (datetime.datetime). Callers that work with inclusive calendar dates
        should use ``parse_datetime_bound(..., upper=True)``.
        For example, the range “2024-01-01 to 2024-01-31” becomes:
          - ``start_date = 2024-01-01 00:00:00``, and
          - ``end_date = 2024-02-01 00:00:00``
    :param cumulative: Flag indicating whether to calculate cumulatively, where
        each month adds to previous totals (bool)

    :returns: "report" dict containing following fields:
        report["StorageName"]: Name of Storage Service queried
        report["Locations"]: List of result locations ordered desc by size
    """
    report = {}
    report[fields.FIELD_STORAGE_NAME] = get_storage_service_name(storage_service_id)

    locations = sorted(
        _get_storage_locations(storage_service_id), key=lambda loc: loc.id
    )
    location_ids = [loc.id for loc in locations]
    days = _get_days_covered_by_date_range(storage_service_id, start_date, end_date)

    # Build a single aggregated query that returns daily metrics per location.
    # MySQL: DATE(AIP.create_date) groups by calendar day.
    day_col = db.func.date(AIP.create_date).label("day")

    # CASE expression to count only original files.
    orig_file_case = case((File.file_type == FileType.original, 1), else_=0)

    aggregated = (
        db.session.query(
            day_col,
            AIP.storage_location_id.label("storage_location_id"),
            db.func.count(db.func.distinct(AIP.id)).label("aips"),
            db.func.coalesce(db.func.sum(File.size), 0).label("size_sum"),
            db.func.coalesce(db.func.sum(orig_file_case), 0).label("orig_files"),
        )
        .outerjoin(File, File.aip_id == AIP.id)
        .filter(AIP.storage_service_id == storage_service_id)
        .filter(AIP.create_date >= start_date)
        .filter(AIP.create_date < end_date)
        .group_by(day_col, AIP.storage_location_id)
        .all()
    )

    # Map aggregated rows for quick lookup.
    daily_loc_metrics = {}  # {day_str: {loc_id: {aips,size,files}}}
    for row in aggregated:
        # row.day can be date/datetime; normalize to YYYY-MM-DD string
        day_str = str(row.day)
        loc_id = row.storage_location_id
        day_bucket = daily_loc_metrics.setdefault(day_str, {})
        day_bucket[loc_id] = {
            fields.FIELD_AIPS: int(row.aips or 0),
            fields.FIELD_SIZE: int(row.size_sum or 0),
            fields.FIELD_FILE_COUNT: int(row.orig_files or 0),
        }

    # Prepare cumulative running totals per location if needed
    running_totals = {
        loc_id: {fields.FIELD_AIPS: 0, fields.FIELD_SIZE: 0, fields.FIELD_FILE_COUNT: 0}
        for loc_id in location_ids
    }

    results = {}
    for day in days:
        daily_locations_data = []

        # Use aggregated metrics for the given day; default to zeros.
        day_metrics = daily_loc_metrics.get(day, {})
        for loc in locations:
            base = day_metrics.get(
                loc.id,
                {
                    fields.FIELD_AIPS: 0,
                    fields.FIELD_SIZE: 0,
                    fields.FIELD_FILE_COUNT: 0,
                },
            )

            if cumulative:
                run = running_totals[loc.id]
                run[fields.FIELD_AIPS] += base[fields.FIELD_AIPS]
                run[fields.FIELD_SIZE] += base[fields.FIELD_SIZE]
                run[fields.FIELD_FILE_COUNT] += base[fields.FIELD_FILE_COUNT]
                aips_val = run[fields.FIELD_AIPS]
                size_val = run[fields.FIELD_SIZE]
                files_val = run[fields.FIELD_FILE_COUNT]
            else:
                aips_val = base[fields.FIELD_AIPS]
                size_val = base[fields.FIELD_SIZE]
                files_val = base[fields.FIELD_FILE_COUNT]

            loc_info = {
                fields.FIELD_ID: loc.id,
                fields.FIELD_UUID: loc.uuid,
                fields.FIELD_STORAGE_LOCATION: loc.description,
                fields.FIELD_AIPS: aips_val,
                fields.FIELD_SIZE: size_val,
                fields.FIELD_FILE_COUNT: files_val,
            }
            daily_locations_data.append(loc_info)

        results[day] = daily_locations_data

    report[fields.FIELD_LOCATIONS_USAGE_OVER_TIME] = results

    return report



import datetime
import uuid

from AIPscan.models import AIP
from AIPscan.models import StorageService


class MockAIPsByFormatOrPUIDResult:
    """Fixture for mocking AIPs by format/PUID results with labels."""

    def __init__(self, mock_id, name, uuid, file_count, total_size):
        self.id = mock_id
        self.name = name
        self.uuid = uuid
        self.file_count = file_count
        self.total_size = total_size


MOCK_AIPS_BY_FORMAT_OR_PUID_QUERY_RESULTS = [
    MockAIPsByFormatOrPUIDResult(
        mock_id=1,
        name="aip0",
        uuid="11111111-1111-1111-1111-111111111111",
        file_count=5,
        total_size=12345678,
    ),
    MockAIPsByFormatOrPUIDResult(
        mock_id=2,
        name="aip1",
        uuid="22222222-2222-2222-2222-222222222222",
        file_count=3,
        total_size=123456,
    ),
    MockAIPsByFormatOrPUIDResult(
        mock_id=5,
        name="aip2",
        uuid="33333333-3333-3333-3333-333333333333",
        file_count=2,
        total_size=12345,
    ),
    MockAIPsByFormatOrPUIDResult(
        mock_id=10,
        name="aip3",
        uuid="44444444-4444-4444-4444-444444444444",
        file_count=1,
        total_size=123,
    ),
]

MOCK_STORAGE_SERVICE_ID = 1
MOCK_STORAGE_SERVICE_NAME = "some name"
MOCK_STORAGE_SERVICE = StorageService(
    name=MOCK_STORAGE_SERVICE_NAME,
    url="http://example.com",
    user_name="test",
    api_key="test",
    download_limit=20,
    download_offset=10,
    default=False,
)
MOCK_STORAGE_SERVICE.id = MOCK_STORAGE_SERVICE_ID

MOCK_STORAGE_SERVICE_ID_2 = 2
MOCK_STORAGE_SERVICE_2 = StorageService(
    name="another name",
    url="http://anotherexample.com",
    user_name="test2",
    api_key="test_api_2",
    download_limit=30,
    download_offset=40,
    default=False,
)
MOCK_STORAGE_SERVICE_2.id = MOCK_STORAGE_SERVICE_ID_2

MOCK_STORAGE_SERVICES = [MOCK_STORAGE_SERVICE, MOCK_STORAGE_SERVICE_2]

MOCK_STORAGE_LOCATION_ID = 1

MOCK_AIP_NAME = "Test transfer"
MOCK_AIP_UUID = str(uuid.uuid4())
MOCK_AIP = AIP(
    uuid=MOCK_AIP_UUID,
    transfer_name=MOCK_AIP_NAME,
    create_date=datetime.datetime.now(),
    mets_sha256="test",
    size=0,
    storage_service_id=MOCK_STORAGE_SERVICE_ID,
    storage_location_id=MOCK_STORAGE_LOCATION_ID,
    fetch_job_id=1,
    origin_pipeline_id=1,
)



from AIPscan.conftest import AIP_1_NAME
from AIPscan.conftest import AIP_1_UUID
from AIPscan.conftest import AIP_2_NAME
from AIPscan.conftest import AIP_2_UUID
from AIPscan.conftest import JPEG_1_01_FORMAT_VERSION
from AIPscan.conftest import JPEG_1_01_PUID
from AIPscan.conftest import JPEG_1_02_FORMAT_VERSION
from AIPscan.conftest import JPEG_1_02_PUID
from AIPscan.conftest import JPEG_FILE_FORMAT
from AIPscan.conftest import ORIGINAL_FILE_1_NAME
from AIPscan.conftest import ORIGINAL_FILE_1_UUID
from AIPscan.conftest import ORIGINAL_FILE_2_NAME
from AIPscan.conftest import ORIGINAL_FILE_2_UUID
from AIPscan.conftest import PRESERVATION_FILE_1_NAME
from AIPscan.conftest import PRESERVATION_FILE_1_UUID
from AIPscan.conftest import PRESERVATION_FILE_2_NAME
from AIPscan.conftest import PRESERVATION_FILE_2_UUID
from AIPscan.conftest import STORAGE_SERVICE_NAME
from AIPscan.conftest import TIFF_FILE_FORMAT
from AIPscan.Data import fields
from AIPscan.Data import report_data

# Two preservation files are created in fixture.
EXPECTED_RESULTS_COUNT = 2

STORAGE_LOCATION_DESCRIPTION = "test storage location"


def test_preservation_derivatives(preservation_derivatives):
    """Test preservation derivatives report_data endpoint."""
    report = report_data.preservation_derivatives(
        storage_service_id=1, storage_location_id=1
    )
    file_data = report[fields.FIELD_FILES]

    assert report[fields.FIELD_STORAGE_NAME] == STORAGE_SERVICE_NAME
    assert report[fields.FIELD_STORAGE_LOCATION] == STORAGE_LOCATION_DESCRIPTION
    assert len(file_data) == EXPECTED_RESULTS_COUNT

    preservation_file_1 = [
        file_
        for file_ in file_data
        if file_[fields.FIELD_UUID] == PRESERVATION_FILE_1_UUID
    ][0]
    preservation_file_2 = [
        file_
        for file_ in file_data
        if file_[fields.FIELD_UUID] == PRESERVATION_FILE_2_UUID
    ][0]

    # Test first preservation file data.
    assert preservation_file_1.get(fields.FIELD_AIP_UUID) == AIP_1_UUID
    assert preservation_file_1.get(fields.FIELD_AIP_NAME) == AIP_1_NAME

    assert preservation_file_1.get(fields.FIELD_UUID) == PRESERVATION_FILE_1_UUID
    assert preservation_file_1.get(fields.FIELD_NAME) == PRESERVATION_FILE_1_NAME
    assert preservation_file_1.get(fields.FIELD_FORMAT) == TIFF_FILE_FORMAT

    assert preservation_file_1.get(fields.FIELD_ORIGINAL_UUID) == ORIGINAL_FILE_1_UUID
    assert preservation_file_1.get(fields.FIELD_ORIGINAL_NAME) == ORIGINAL_FILE_1_NAME
    assert preservation_file_1.get(fields.FIELD_ORIGINAL_FORMAT) == JPEG_FILE_FORMAT
    assert (
        preservation_file_1.get(fields.FIELD_ORIGINAL_VERSION)
        == JPEG_1_01_FORMAT_VERSION
    )
    assert preservation_file_1.get(fields.FIELD_ORIGINAL_PUID) == JPEG_1_01_PUID

    # Test second preservation file data.
    assert preservation_file_2.get(fields.FIELD_AIP_UUID) == AIP_2_UUID
    assert preservation_file_2.get(fields.FIELD_AIP_NAME) == AIP_2_NAME

    assert preservation_file_2.get(fields.FIELD_UUID) == PRESERVATION_FILE_2_UUID
    assert preservation_file_2.get(fields.FIELD_NAME) == PRESERVATION_FILE_2_NAME
    assert preservation_file_2.get(fields.FIELD_FORMAT) == TIFF_FILE_FORMAT

    assert preservation_file_2.get(fields.FIELD_ORIGINAL_UUID) == ORIGINAL_FILE_2_UUID
    assert preservation_file_2.get(fields.FIELD_ORIGINAL_NAME) == ORIGINAL_FILE_2_NAME
    assert preservation_file_2.get(fields.FIELD_ORIGINAL_FORMAT) == JPEG_FILE_FORMAT
    assert (
        preservation_file_2.get(fields.FIELD_ORIGINAL_VERSION)
        == JPEG_1_02_FORMAT_VERSION
    )
    assert preservation_file_2.get(fields.FIELD_ORIGINAL_PUID) == JPEG_1_02_PUID
